from sonusai.mixture.datatypes import Feature
from sonusai.mixture.datatypes import SpectralMask


def apply_spectral_mask(feature: Feature, spectral_mask: SpectralMask, seed: int = None) -> Feature:
    """Apply frequency and time masking

    Implementation of SpecAugment: A Simple Data Augmentation Method for Automatic Speech Recognition

    Ref: https://arxiv.org/pdf/1904.08779.pdf

    f_width consecutive bands [f_start, f_start + f_width) are masked, where f_width is chosen from a uniform
    distribution from 0 to the f_max_width, and f_start is chosen from [0, bands - f_width).

    t_width consecutive frames [t_start, t_start + t_width) are masked, where t_width is chosen from a uniform
    distribution from 0 to the t_max_width, and t_start is chosen from [0, frames - t_width).

    A time mask cannot be wider than t_max_percent times the number of frames.

    :param feature: Numpy array of feature data [frames, strides, bands]
    :param spectral_mask: Spectral mask parameters
    :param seed: Random number seed
    :return: Augmented feature
    """
    import numpy as np

    from sonusai import SonusAIError

    if feature.ndim != 3:
        raise SonusAIError('feature input must have three dimensions [frames, strides, bands]')

    frames, strides, bands = feature.shape

    f_max_width = spectral_mask.f_max_width
    if f_max_width not in range(0, bands + 1):
        f_max_width = bands

    rng = np.random.default_rng(seed)

    # apply f_num frequency masks to the feature
    for _ in range(spectral_mask.f_num):
        f_width = int(rng.uniform(0, f_max_width))
        f_start = rng.integers(0, bands - f_width, endpoint=True)
        feature[:, :, f_start:f_start + f_width] = 0

    # apply t_num time masks to the feature
    t_upper_bound = int(spectral_mask.t_max_percent / 100 * frames)
    for _ in range(spectral_mask.t_num):
        t_width = min(int(rng.uniform(0, spectral_mask.t_max_width)), t_upper_bound)
        t_start = rng.integers(0, frames - t_width, endpoint=True)
        feature[t_start:t_start + t_width, :, :] = 0

    return feature



from pyaaware import ForwardTransform

from sonusai.mixture.datatypes import AudioF
from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Truth
from sonusai.mixture.truth_functions.data import Data


def target_f(data: Data) -> Truth:
    """Frequency domain target truth function

Calculates the true transform of the target using the STFT
configuration defined by the feature. This will include a
forward transform window if defined by the feature.

Output shape: [:, num_classes]
                    (target stacked real, imag; or real only for tdac-co)
    """

    from sonusai import SonusAIError

    if data.config.num_classes != data.num_bands:
        raise SonusAIError(f'Invalid num_classes for target_f truth: {data.config.num_classes}')

    target_freq = _execute_fft(data.target_audio, data.target_fft, len(data.offsets))
    for idx, offset in enumerate(data.offsets):
        data.truth = _stack_real_imag(data=target_freq[idx],
                                      offset=offset,
                                      frame_size=data.frame_size,
                                      zero_based_indices=data.zero_based_indices,
                                      bins=data.target_fft.bins,
                                      ttype=data.ttype,
                                      start=0,
                                      truth=data.truth)

    return data.truth


# TODO: Need Data to include mixture audio to do this correctly
def target_mixture_f(data: Data) -> Truth:
    """Frequency domain target and mixture truth function

Calculates the true transform of the target and the mixture
using the STFT configuration defined by the feature. This
will include a forward transform window if defined by the
feature.

Output shape: [:, 2 * num_classes]
                    (target stacked real, imag; or real only for tdac-co)
                    (mixture stacked real, imag; or real only for tdac-co)
    """
    from sonusai import SonusAIError

    if data.config.num_classes != 2 * data.num_bands:
        raise SonusAIError(f'Invalid num_classes for target_mixture_f truth: {data.config.num_classes}')

    target_freq = _execute_fft(data.target_audio, data.target_fft, len(data.offsets))
    mixture_freq = _execute_fft(data.mixture_audio, data.mixture_fft, len(data.offsets))

    for idx, offset in enumerate(data.offsets):
        data.truth = _stack_real_imag(data=target_freq[idx],
                                      offset=offset,
                                      frame_size=data.frame_size,
                                      zero_based_indices=data.zero_based_indices,
                                      bins=data.target_fft.bins,
                                      ttype=data.ttype,
                                      start=0,
                                      truth=data.truth)

        data.truth = _stack_real_imag(data=mixture_freq[idx],
                                      offset=offset,
                                      frame_size=data.frame_size,
                                      zero_based_indices=data.zero_based_indices,
                                      bins=data.target_fft.bins,
                                      ttype=data.ttype,
                                      start=data.target_fft.bins * 2,
                                      truth=data.truth)

    return data.truth


def target_swin_f(data: Data) -> Truth:
    """Frequency domain target with synthesis window truth function

Calculates the true transform of the target using the STFT
configuration defined by the feature. This will include a
forward transform window if defined by the feature and also
the inverse transform (or synthesis) window.

Output shape: [:, 2 * bins] (stacked real, imag)
    """
    import numpy as np

    from sonusai import SonusAIError

    if data.config.num_classes != 2 * data.target_fft.bins:
        raise SonusAIError(f'Invalid num_classes for target_swin_f truth: {data.config.num_classes}')

    for idx, offset in enumerate(data.offsets):
        target_freq, _ = data.target_fft.execute(
            np.multiply(data.target_audio[offset:offset + data.frame_size], data.swin))
        target_freq = target_freq.transpose()

        indices = slice(offset, offset + data.frame_size)
        for index in data.zero_based_indices:
            bins = _get_bin_slice(index, data.target_fft.bins)
            data.truth[indices, bins] = np.real(target_freq[idx])

            bins = _get_bin_slice(bins.stop, data.target_fft.bins)
            data.truth[indices, bins] = np.imag(target_freq[idx])

    return data.truth


def _execute_fft(audio: AudioT, transform: ForwardTransform, expected_frames: int) -> AudioF:
    from sonusai import SonusAIError

    freq, _ = transform.execute_all(audio)
    freq = freq.transpose()
    if len(freq) != expected_frames:
        raise SonusAIError(f'Number of frames, {len(freq)}, is not number of frames expected, {expected_frames}')
    return freq


def _get_bin_slice(start: int, length: int) -> slice:
    return slice(start, start + length)


def _stack_real_imag(data: AudioF,
                     offset: int,
                     frame_size: int,
                     zero_based_indices: list[int],
                     bins: int,
                     ttype: str,
                     start: int,
                     truth: Truth) -> Truth:
    import numpy as np

    i = _get_bin_slice(offset, frame_size)
    for index in zero_based_indices:
        b = _get_bin_slice(index + start, bins)
        truth[i, b] = np.real(data)

        if ttype != 'tdac-co':
            b = _get_bin_slice(b.stop, bins)
            truth[i, b] = np.imag(data)

    return truth



from sonusai.utils.asr_manifest_functions import PathInfo
from sonusai.utils.asr_manifest_functions import TranscriptData


def collect_librispeech_transcripts(paths: list[str] | str) -> TranscriptData:
    from glob import iglob
    from os.path import abspath
    from os.path import dirname
    from os.path import join

    from sonusai import SonusAIError

    entries: TranscriptData = {}
    if not isinstance(paths, list):
        paths = [paths]

    for p in paths:
        location = join(abspath(p), '**', '*.trans.txt')
        for file in iglob(pathname=location, recursive=True):
            root = dirname(file)
            with open(file, encoding='utf-8') as f:
                for line in f:
                    name, text = line[: line.index(' ')], line[line.index(' ') + 1:]
                    name = join(root, name)
                    if name in entries:
                        raise SonusAIError(f'{name} already exists in transcript data')
                    entries[name] = text.lower().strip()
    return entries


def get_librispeech_manifest_entry(entry: PathInfo, transcript_data: TranscriptData) -> dict:
    from os.path import splitext
    from subprocess import check_output

    from sonusai import SonusAIError

    name = splitext(entry.abs_path)[0]
    duration = float(check_output(f'soxi -D {entry.abs_path}', shell=True))
    if name not in transcript_data.keys():
        raise SonusAIError(f'Could not find {name} in transcript data')

    return {
        'audio_filepath': entry.audio_filepath,
        'text':           transcript_data[name],
        'duration':       duration,
    }



from typing import Callable

from sonusai.mixture import ClassCount
from sonusai.mixture import MixtureDatabase


def print_mixture_details(mixdb: MixtureDatabase,
                          mixid: int = None,
                          desc_len: int = 1,
                          print_fn: Callable = print) -> None:
    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.utils import seconds_to_hms

    if mixid is not None:
        if 0 < mixid >= mixdb.num_mixtures:
            raise SonusAIError(f'Given mixid is outside valid range of 0:{mixdb.num_mixtures - 1}.')

        print_fn(f'Mixture {mixid} details')
        mixture = mixdb.mixture(mixid)
        target_files = [mixdb.target_files[target.file_id] for target in mixture.targets]
        target_augmentations = [target.augmentation for target in mixture.targets]
        noise_file = mixdb.noise_file(mixture.noise.file_id)
        for t_idx, target_file in enumerate(target_files):
            print_fn(f'  Target {t_idx}')
            print_fn(f'{"    Name":{desc_len}} {target_file.name}')
            print_fn(f'{"    Duration":{desc_len}} {seconds_to_hms(target_file.duration)}')
            for ts_idx, truth_setting in enumerate(target_file.truth_settings):
                print_fn(f'    Truth setting {ts_idx}')
                print_fn(f'{"      Index":{desc_len}} {truth_setting.index}')
                print_fn(f'{"      Function":{desc_len}} {truth_setting.function}')
                print_fn(f'{"      Config":{desc_len}} {truth_setting.config}')
            print_fn(f'{"    Augmentation":{desc_len}} {target_augmentations[t_idx]}')
        print_fn(f'{"  Samples":{desc_len}} {mixture.samples}')
        print_fn(f'{"  Feature frames":{desc_len}} {mixdb.mixture_feature_frames(mixid)}')
        print_fn(f'{"  Noise file":{desc_len}} {noise_file.name}')
        noise_offset_percent = int(np.round(100 * mixture.noise.offset / float(noise_file.duration * SAMPLE_RATE)))
        print_fn(f'{"  Noise offset":{desc_len}} {mixture.noise.offset} samples ({noise_offset_percent}%)')
        print_fn(f'{"  SNR":{desc_len}} {mixture.snr} dB{" (random)" if mixture.snr.is_random else ""}')
        print_fn(f'{"  Target gain":{desc_len}} {[target.gain for target in mixture.targets]}')
        print_fn(f'{"  Target SNR gain":{desc_len}} {mixture.target_snr_gain}')
        print_fn(f'{"  Noise SNR gain":{desc_len}} {mixture.noise_snr_gain}')
        print_fn('')


def print_class_count(class_count: ClassCount,
                      length: int,
                      print_fn: Callable = print,
                      all_class_counts: bool = False) -> None:
    from sonusai.utils import max_text_width

    print_fn(f'Class count:')
    idx_len = max_text_width(len(class_count))
    for idx, count in enumerate(class_count):
        if all_class_counts or count > 0:
            desc = f'  class {idx + 1:{idx_len}}'
            print_fn(f'{desc:{length}} {count}')



from dataclasses import dataclass


@dataclass(frozen=True)
class WerResult:
    wer: float
    words: int
    substitutions: float
    deletions: float
    insertions: float


def calc_wer(hypothesis: list[str] | str, reference: list[str] | str) -> WerResult:
    """Computes average word error rate between two texts represented as corresponding strings or lists of strings.

    :param hypothesis: the hypothesis sentence(s) as a string or list of strings
    :param reference: the reference sentence(s) as a string or list of strings
    :return: a WerResult object with error, words, insertions, deletions, substitutions
    """
    import jiwer

    transformation = jiwer.Compose([
        jiwer.ToLowerCase(),
        jiwer.RemovePunctuation(),
        jiwer.RemoveWhiteSpace(replace_by_space=True),
        jiwer.RemoveMultipleSpaces(),
        jiwer.Strip(),
        jiwer.RemoveEmptyStrings(),
        jiwer.ReduceToListOfListOfWords(word_delimiter=' ')
    ])

    if isinstance(reference, str):
        reference = [reference]
    if isinstance(hypothesis, str):
        hypothesis = [hypothesis]

    # jiwer does not allow empty string
    measures = {'insertions':    0,
                'substitutions': 0,
                'deletions':     0,
                'hits':          0}
    if any(len(t) == 0 for t in reference):
        if any(len(t) != 0 for t in hypothesis):
            measures['insertions'] = len(hypothesis)
    else:
        measures = jiwer.compute_measures(truth=reference,
                                          hypothesis=hypothesis,
                                          truth_transform=transformation,
                                          hypothesis_transform=transformation)

    errors = measures['substitutions'] + measures['deletions'] + measures['insertions']
    words = measures['hits'] + measures['substitutions'] + measures['deletions']

    if words != 0:
        wer = errors / words
        substitutions_rate = measures['substitutions'] / words
        deletions_rate = measures['deletions'] / words
        insertions_rate = measures['insertions'] / words
    else:
        wer = float('inf')
        substitutions_rate = float('inf')
        deletions_rate = float('inf')
        insertions_rate = float('inf')

    return WerResult(wer=wer,
                     words=int(words),
                     substitutions=substitutions_rate,
                     deletions=deletions_rate,
                     insertions=insertions_rate)



from functools import lru_cache

from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from sonusai.mixture.datatypes import AudioF
from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import EnergyT
from sonusai.mixture.datatypes import ImpulseResponseData


def get_next_noise(audio: AudioT, offset: int, length: int) -> AudioT:
    """Get next sequence of noise data from noise audio

    :param audio: Overall noise audio (entire file's worth of data)
    :param offset: Starting sample
    :param length: Number of samples to get
    :return: Sequence of noise audio data
    """
    import numpy as np

    return np.take(audio, range(offset, offset + length), mode='wrap')


def calculate_transform_from_audio(audio: AudioT,
                                   transform: ForwardTransform) -> tuple[AudioF, EnergyT]:
    """Apply forward transform to input audio data to generate transform data

    :param audio: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain data [frames, bins], Energy [frames]
    """
    f, e = transform.execute_all(audio)
    return f.transpose(), e


def calculate_audio_from_transform(data: AudioF,
                                   transform: InverseTransform,
                                   trim: bool = True) -> tuple[AudioT, EnergyT]:
    """Apply inverse transform to input transform data to generate audio data

    :param data: Frequency domain data [frames, bins]
    :param transform: InverseTransform object
    :param trim: Removes starting samples so output waveform will be time-aligned with input waveform to the transform
    :return: Time domain data [samples], Energy [frames]
    """
    t, e = transform.execute_all(data.transpose())
    if trim:
        t = t[transform.N - transform.R:]

    return t, e


def get_duration(audio: AudioT) -> float:
    """Get duration of audio in seconds

    :param audio: Time domain data [samples]
    :return: Duration of audio in seconds
    """
    from .constants import SAMPLE_RATE

    return len(audio) / SAMPLE_RATE


def validate_input_file(input_filepath: str) -> None:
    from os.path import exists
    from os.path import splitext

    from soundfile import available_formats

    from sonusai import SonusAIError

    if not exists(input_filepath):
        raise SonusAIError(f'input_filepath {input_filepath} does not exist.')

    ext = splitext(input_filepath)[1][1:].lower()
    read_formats = [item.lower() for item in available_formats().keys()]
    if ext not in read_formats:
        raise SonusAIError(f'This installation cannot process .{ext} files')


@lru_cache
def read_audio(name: str) -> AudioT:
    """Read audio data from a file

    :param name: File name
    :return: Array of time domain audio data
    """
    from .soundfile_audio import read_audio

    return read_audio(name)


@lru_cache
def read_ir(name: str) -> ImpulseResponseData:
    """Read impulse response data

    :param name: File name
    :return: ImpulseResponseData object
    """
    from .soundfile_audio import read_ir

    return read_ir(name)


@lru_cache
def get_num_samples(name: str) -> int:
    """Get the number of samples resampled to the SonusAI sample rate in the given file

    :param name: File name
    :return: number of samples in resampled audio
    """
    from .soundfile_audio import get_num_samples

    return get_num_samples(name)



from dataclasses import dataclass
from typing import Any
from typing import Optional

from sonusai.mixture.datatypes import AudioT


@dataclass(frozen=True)
class Data:
    audio: AudioT
    whisper_model: Optional[Any] = None
    whisper_model_name: Optional[str] = None
    device: Optional[str] = None
    cpu_threads: Optional[int] = None
    compute_type: Optional[str] = None
    beam_size: Optional[int] = None



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Feature


def get_feature_from_audio(audio: AudioT, feature: str) -> Feature:
    from dataclasses import asdict

    import numpy as np
    from pyaaware import FeatureGenerator

    from .augmentation import pad_audio_to_frame
    from .datatypes import FeatureGeneratorConfig
    from .datatypes import TransformConfig
    from .helpers import forward_transform
    from .truth import truth_reduction

    num_classes = 1
    truth_mutex = False
    truth_reduction_function = 'max'

    fg_config = FeatureGeneratorConfig(feature_mode=feature,
                                       num_classes=num_classes,
                                       truth_mutex=truth_mutex)
    fg = FeatureGenerator(**asdict(fg_config))
    feature_step_samples = fg.ftransform_R * fg.decimation * fg.step

    audio = pad_audio_to_frame(audio, feature_step_samples)
    samples = len(audio)
    audio_f = forward_transform(audio, TransformConfig(N=fg.ftransform_N,
                                                       R=fg.ftransform_R,
                                                       bin_start=fg.bin_start,
                                                       bin_end=fg.bin_end,
                                                       ttype=fg.ftransform_ttype))

    transform_frames = samples // fg.ftransform_R
    feature_frames = samples // feature_step_samples

    truth_t = np.empty((samples, num_classes), dtype=np.float32)

    data = np.empty((feature_frames, fg.stride, fg.num_bands), dtype=np.float32)

    feature_frame = 0
    for transform_frame in range(transform_frames):
        indices = slice(transform_frame * fg.ftransform_R, (transform_frame + 1) * fg.ftransform_R)
        fg.execute(audio_f[transform_frame], truth_reduction(truth_t[indices], truth_reduction_function))

        if fg.eof():
            data[feature_frame] = fg.feature()
            feature_frame += 1

    return data



from typing import Any

from sonusai.mixture.datatypes import AudioF
from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import AudiosT
from sonusai.mixture.datatypes import Augmentation
from sonusai.mixture.datatypes import AugmentationRules
from sonusai.mixture.datatypes import Augmentations
from sonusai.mixture.datatypes import Feature
from sonusai.mixture.datatypes import FeatureGeneratorConfig
from sonusai.mixture.datatypes import FeatureGeneratorInfo
from sonusai.mixture.datatypes import GeneralizedIDs
from sonusai.mixture.datatypes import Mixture
from sonusai.mixture.datatypes import NoiseFile
from sonusai.mixture.datatypes import NoiseFiles
from sonusai.mixture.datatypes import Segsnr
from sonusai.mixture.datatypes import Target
from sonusai.mixture.datatypes import TargetFiles
from sonusai.mixture.datatypes import Targets
from sonusai.mixture.datatypes import TransformConfig
from sonusai.mixture.datatypes import Truth
from sonusai.mixture.mixdb import MixtureDatabase


def generic_ids_to_list(num_ids: int, ids: GeneralizedIDs = None) -> list[int]:
    """Resolve generalized IDs to a list of integers

    :param num_ids: Total number of indices
    :param ids: Generalized IDs
    :return: List of ID integers
    """
    from sonusai import SonusAIError

    all_ids = list(range(num_ids))

    if ids is None:
        return all_ids

    if isinstance(ids, str):
        if ids == '*':
            return all_ids

        try:
            result = eval(f'{all_ids}[{ids}]')
            if not isinstance(result, list):
                result = [result]
            return result
        except NameError:
            raise SonusAIError(f'Empty ids {ids}')

    if isinstance(ids, range):
        result = list(ids)
    elif isinstance(ids, int):
        result = [ids]
    else:
        result = ids

    if not all(isinstance(x, int) and 0 <= x < num_ids for x in result):
        raise SonusAIError(f'Invalid entries in ids of {ids}')

    if not result:
        raise SonusAIError(f'Empty ids {ids}')

    return result


def get_feature_generator_info(fg_config: FeatureGeneratorConfig) -> FeatureGeneratorInfo:
    from dataclasses import asdict

    from pyaaware import FeatureGenerator

    from .datatypes import FeatureGeneratorInfo
    from .datatypes import TransformConfig

    fg = FeatureGenerator(**asdict(fg_config))

    return FeatureGeneratorInfo(
        decimation=fg.decimation,
        stride=fg.stride,
        step=fg.step,
        num_bands=fg.num_bands,
        ft_config=TransformConfig(N=fg.ftransform_N,
                                  R=fg.ftransform_R,
                                  bin_start=fg.bin_start,
                                  bin_end=fg.bin_end,
                                  ttype=fg.ftransform_ttype),
        eft_config=TransformConfig(N=fg.eftransform_N,
                                   R=fg.eftransform_R,
                                   bin_start=fg.bin_start,
                                   bin_end=fg.bin_end,
                                   ttype=fg.eftransform_ttype),
        it_config=TransformConfig(N=fg.itransform_N,
                                  R=fg.itransform_R,
                                  bin_start=fg.bin_start,
                                  bin_end=fg.bin_end,
                                  ttype=fg.itransform_ttype)
    )


def write_mixture_data(mixdb: MixtureDatabase,
                       mixture: Mixture,
                       items: list[tuple[str, Any]] | tuple[str, Any]) -> None:
    """Write mixture data to a mixture HDF5 file

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param items: Tuple(s) of (name, data)
    """
    import h5py

    if not isinstance(items, list):
        items = [items]

    name = mixdb.location_filename(mixture.name)
    with h5py.File(name=name, mode='a') as f:
        for item in items:
            if item[0] in f:
                del f[item[0]]
            f.create_dataset(name=item[0], data=item[1])


def mixture_metadata(mixdb: MixtureDatabase, mixture: Mixture) -> str:
    """Create a string of metadata for a Mixture

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :return: String of metadata
    """
    metadata = ''
    for mi, target in enumerate(mixture.targets):
        target_file = mixdb.target_file(target.file_id)
        target_augmentation = target.augmentation
        metadata += f'target {mi} name: {target_file.name}\n'
        metadata += f'target {mi} augmentation: {target.augmentation.to_dict()}\n'
        if target_augmentation.ir is None:
            ir_name = None
        else:
            ir_name = mixdb.impulse_response_file(target_augmentation.ir)
        metadata += f'target {mi} ir: {ir_name}\n'
        metadata += f'target {mi} target_gain: {target.gain}\n'
        truth_settings = target_file.truth_settings
        for tsi in range(len(truth_settings)):
            metadata += f'target {mi} truth index {tsi}: {truth_settings[tsi].index}\n'
            metadata += f'target {mi} truth function {tsi}: {truth_settings[tsi].function}\n'
            metadata += f'target {mi} truth config {tsi}: {truth_settings[tsi].config}\n'
        metadata += f'target {mi} asr: {mixdb.target_asr_data(target.file_id)}\n'
    noise = mixdb.noise_file(mixture.noise.file_id)
    noise_augmentation = mixture.noise.augmentation
    metadata += f'noise name: {noise.name}\n'
    metadata += f'noise augmentation: {noise_augmentation.to_dict()}\n'
    if noise_augmentation.ir is None:
        ir_name = None
    else:
        ir_name = mixdb.impulse_response_file(noise_augmentation.ir)
    metadata += f'noise ir: {ir_name}\n'
    metadata += f'noise offset: {mixture.noise.offset}\n'
    metadata += f'snr: {mixture.snr}\n'
    metadata += f'random_snr: {mixture.snr.is_random}\n'
    metadata += f'samples: {mixture.samples}\n'
    metadata += f'target_snr_gain: {float(mixture.target_snr_gain)}\n'
    metadata += f'noise_snr_gain: {float(mixture.noise_snr_gain)}\n'

    return metadata


def write_mixture_metadata(mixdb: MixtureDatabase, mixture: Mixture) -> None:
    """Write mixture metadata to a text file

    :param mixdb: Mixture database
    :param mixture: Mixture record
    """
    from os.path import splitext

    name = mixdb.location_filename(splitext(mixture.name)[0] + '.txt')
    with open(file=name, mode='w') as f:
        f.write(mixture_metadata(mixdb, mixture))


def from_mixture(mixture: Mixture) -> tuple[str, int, str, int, float, bool, float, int, int, int, float]:
    return (mixture.name,
            mixture.noise.file_id,
            mixture.noise.augmentation.to_json(),
            mixture.noise.offset,
            mixture.noise_snr_gain,
            mixture.snr.is_random,
            mixture.snr,
            mixture.samples,
            mixture.spectral_mask_id,
            mixture.spectral_mask_seed,
            mixture.target_snr_gain)


def to_mixture(entry: tuple[str, int, str, int, float, bool, float, int, int, int, float], targets: Targets) -> Mixture:
    import json

    from sonusai.utils import dataclass_from_dict
    from .datatypes import Augmentation
    from .datatypes import Mixture
    from .datatypes import Noise
    from .datatypes import UniversalSNR

    return Mixture(targets=targets,
                   name=entry[0],
                   noise=Noise(file_id=entry[1],
                               augmentation=dataclass_from_dict(Augmentation, json.loads(entry[2])),
                               offset=entry[3]),
                   noise_snr_gain=entry[4],
                   snr=UniversalSNR(is_random=entry[5], value=entry[6]),
                   samples=entry[7],
                   spectral_mask_id=entry[8],
                   spectral_mask_seed=entry[9],
                   target_snr_gain=entry[10])


def from_target(target: Target) -> tuple[int, str, float]:
    return target.file_id, target.augmentation.to_json(), target.gain


def to_target(entry: tuple[int, str, float]) -> Target:
    import json

    from sonusai.utils import dataclass_from_dict
    from .datatypes import Augmentation
    from .datatypes import Target

    return Target(file_id=entry[0],
                  augmentation=dataclass_from_dict(Augmentation, json.loads(entry[1])),
                  gain=entry[2])


def read_mixture_data(name: str, items: list[str] | str) -> Any:
    """Read mixture data from a mixture HDF5 file

    :param name: Mixture file name
    :param items: String(s) of dataset(s) to retrieve
    :return: Data (or tuple of data)
    """
    from os.path import exists
    from typing import Any

    import h5py
    import numpy as np

    from sonusai import SonusAIError

    def _get_dataset(file: h5py.File, d_name: str) -> Any:
        if d_name in file:
            return np.array(file[d_name])
        return None

    if not isinstance(items, list):
        items = [items]

    if exists(name):
        try:
            with h5py.File(name, 'r') as f:
                result = ([_get_dataset(f, item) for item in items])
        except Exception as e:
            raise SonusAIError(f'Error reading {name}: {e}')
    else:
        result = ([None for _ in items])

    if len(items) == 1:
        result = result[0]

    return result


def get_truth_t(mixdb: MixtureDatabase,
                mixture: Mixture,
                targets_audio: AudiosT,
                noise_audio: AudioT,
                mixture_audio: AudioT) -> Truth:
    """Get the truth_t data for the given mixture record

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param targets_audio: List of augmented target audio data (one per target in the mixup) for the given mixture ID
    :param noise_audio: Augmented noise audio data for the given mixture ID
    :param mixture_audio: Mixture audio data for the given mixture ID
    :return: truth_t data
    """
    import numpy as np

    from sonusai import SonusAIError
    from .datatypes import TruthFunctionConfig
    from .truth import truth_function

    if not all(len(target) == mixture.samples for target in targets_audio):
        raise SonusAIError('Lengths of targets do not match length of mixture')

    if len(noise_audio) != mixture.samples:
        raise SonusAIError('Length of noise does not match length of mixture')

    # TODO: Need to understand how to do this correctly for mixup and target_mixture_f truth
    truth_t = np.zeros((mixture.samples, mixdb.num_classes), dtype=np.float32)
    for idx in range(len(targets_audio)):
        for truth_setting in mixdb.target_file(mixture.targets[idx].file_id).truth_settings:
            config = TruthFunctionConfig(
                feature=mixdb.feature,
                index=truth_setting.index,
                function=truth_setting.function,
                config=truth_setting.config,
                num_classes=mixdb.num_classes,
                mutex=mixdb.truth_mutex,
                target_gain=mixture.targets[idx].gain * mixture.target_snr_gain
            )
            truth_t += truth_function(target_audio=targets_audio[idx],
                                      noise_audio=noise_audio,
                                      mixture_audio=mixture_audio,
                                      config=config)

    return truth_t


def get_ft(mixdb: MixtureDatabase, mixture: Mixture, mixture_audio: AudioT, truth_t: Truth) -> tuple[Feature, Truth]:
    """Get the feature and truth_f data for the given mixture record

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param mixture_audio: Mixture audio data for the given mixid
    :param truth_t: truth_t for the given mixid
    :return: Tuple of (feature, truth_f) data
    """
    from dataclasses import asdict

    import numpy as np
    from pyaaware import FeatureGenerator

    from .spectral_mask import apply_spectral_mask
    from .truth import truth_reduction

    mixture_f = get_mixture_f(mixdb=mixdb, mixture_audio=mixture_audio)

    transform_frames = mixdb.mixture_transform_frames(mixture.samples)
    feature_frames = mixdb.mixture_feature_frames(mixture.samples)

    feature = np.empty((feature_frames, mixdb.fg_stride, mixdb.fg_num_bands), dtype=np.float32)
    truth_f = np.empty((feature_frames, mixdb.num_classes), dtype=np.complex64)

    fg = FeatureGenerator(**asdict(mixdb.fg_config))
    feature_frame = 0
    for transform_frame in range(transform_frames):
        indices = slice(transform_frame * mixdb.ft_config.R, (transform_frame + 1) * mixdb.ft_config.R)
        fg.execute(mixture_f[transform_frame],
                   truth_reduction(truth_t[indices], mixdb.truth_reduction_function))

        if fg.eof():
            feature[feature_frame] = fg.feature()
            truth_f[feature_frame] = fg.truth()
            feature_frame += 1

    if mixture.spectral_mask_id is not None:
        feature = apply_spectral_mask(feature=feature,
                                      spectral_mask=mixdb.spectral_mask(mixture.spectral_mask_id),
                                      seed=mixture.spectral_mask_seed)

    if np.isreal(truth_f).all():
        return feature, truth_f.real

    return feature, truth_f  # type: ignore


def get_segsnr(mixdb: MixtureDatabase, mixture: Mixture, target_audio: AudioT, noise: AudioT) -> Segsnr:
    """Get the segsnr data for the given mixture record

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param target_audio: Augmented target audio data
    :param noise: Augmented noise audio data
    :return: segsnr data
    """
    segsnr_t = get_segsnr_t(mixdb=mixdb, mixture=mixture, target_audio=target_audio, noise_audio=noise)
    return segsnr_t[0::mixdb.ft_config.R]


def get_segsnr_t(mixdb: MixtureDatabase, mixture: Mixture, target_audio: AudioT, noise_audio: AudioT) -> Segsnr:
    """Get the segsnr_t data for the given mixture record

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param target_audio: Augmented target audio data
    :param noise_audio: Augmented noise audio data
    :return: segsnr_t data
    """
    import numpy as np
    from pyaaware import AawareForwardTransform

    from sonusai import SonusAIError

    fft = AawareForwardTransform(N=mixdb.ft_config.N,
                                 R=mixdb.ft_config.R,
                                 bin_start=mixdb.ft_config.bin_start,
                                 bin_end=mixdb.ft_config.bin_end,
                                 ttype=mixdb.ft_config.ttype)

    segsnr_t = np.empty(mixture.samples, dtype=np.float32)

    _, target_energy = fft.execute_all(target_audio)
    _, noise_energy = fft.execute_all(noise_audio)

    offsets = range(0, mixture.samples, mixdb.ft_config.R)
    if len(target_energy) != len(offsets):
        raise SonusAIError(f'Number of frames in energy, {len(target_energy)},'
                           f' is not number of frames in mixture, {len(offsets)}')

    for idx, offset in enumerate(offsets):
        indices = slice(offset, offset + mixdb.ft_config.R)

        if noise_energy[idx] == 0:
            snr = np.float32(np.inf)
        else:
            snr = np.float32(target_energy[idx] / noise_energy[idx])

        segsnr_t[indices] = snr

    return segsnr_t


def get_target(mixdb: MixtureDatabase, mixture: Mixture, targets_audio: AudiosT) -> AudioT:
    """Get the augmented target audio data for the given mixture record

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param targets_audio: List of augmented target audio data (one per target in the mixup)
    :return: Sum of augmented target audio data
    """
    # Apply impulse responses to targets
    import numpy as np

    from .audio import read_ir
    from .augmentation import apply_impulse_response

    targets_ir = []
    for idx, target in enumerate(targets_audio):
        ir_idx = mixture.targets[idx].augmentation.ir
        if ir_idx is not None:
            targets_ir.append(apply_impulse_response(audio=target,
                                                     ir=read_ir(mixdb.impulse_response_file(int(ir_idx)))))
        else:
            targets_ir.append(target)

    # Return sum of targets
    return np.sum(targets_ir, axis=0)


def get_mixture_f(mixdb: MixtureDatabase, mixture_audio: AudioT) -> AudioF:
    """Get the mixture transform for the given mixture

    :param mixdb: Mixture database
    :param mixture_audio: Mixture audio data for the given mixid
    :return: Mixture transform data
    """
    return forward_transform(mixture_audio, mixdb.ft_config)


def forward_transform(audio: AudioT, config: TransformConfig) -> AudioF:
    """Transform time domain data into frequency domain using the forward transform config from the feature

    A new transform is used for each call; i.e., state is not maintained between calls to forward_transform().

    :param audio: Time domain data [samples]
    :param config: Transform configuration
    :return: Frequency domain data [frames, bins]
    """
    from pyaaware import AawareForwardTransform

    from .audio import calculate_transform_from_audio

    audio_f, _ = calculate_transform_from_audio(audio=audio,
                                                transform=AawareForwardTransform(N=config.N,
                                                                                 R=config.R,
                                                                                 bin_start=config.bin_start,
                                                                                 bin_end=config.bin_end,
                                                                                 ttype=config.ttype))
    return audio_f


def inverse_transform(transform: AudioF, config: TransformConfig, trim: bool = True) -> AudioT:
    """Transform frequency domain data into time domain using the inverse transform config from the feature

    A new transform is used for each call; i.e., state is not maintained between calls to inverse_transform().

    :param transform: Frequency domain data [frames, bins]
    :param config: Transform configuration
    :param trim: Removes starting samples so output waveform will be time-aligned with input waveform to the
                 transform
    :return: Time domain data [samples]
    """
    import numpy as np
    from pyaaware import AawareInverseTransform

    from .audio import calculate_audio_from_transform

    audio, _ = calculate_audio_from_transform(data=transform,
                                              transform=AawareInverseTransform(N=config.N,
                                                                               R=config.R,
                                                                               bin_start=config.bin_start,
                                                                               bin_end=config.bin_end,
                                                                               ttype=config.ttype,
                                                                               gain=np.float32(1)),
                                              trim=trim)
    return audio


def check_audio_files_exist(mixdb: MixtureDatabase) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist
    """
    from os.path import exists

    from sonusai import SonusAIError
    from .tokenized_shell_vars import tokenized_expand

    for noise in mixdb.noise_files:
        file_name, _ = tokenized_expand(noise.name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')

    for target in mixdb.target_files:
        file_name, _ = tokenized_expand(target.name)
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')


def augmented_target_samples(target_files: TargetFiles,
                             target_augmentations: AugmentationRules,
                             feature_step_samples: int) -> int:
    from itertools import product

    from .augmentation import estimate_augmented_length_from_length

    target_ids = list(range(len(target_files)))
    target_augmentation_ids = list(range(len(target_augmentations)))
    it = list(product(*[target_ids, target_augmentation_ids]))
    return sum([estimate_augmented_length_from_length(
        length=target_files[fi].samples,
        tempo=target_augmentations[ai].tempo,
        frame_length=feature_step_samples) for fi, ai, in it])


def augmented_noise_samples(noise_files: NoiseFiles, noise_augmentations: Augmentations) -> int:
    from itertools import product

    noise_ids = list(range(len(noise_files)))
    noise_augmentation_ids = list(range(len(noise_augmentations)))
    it = list(product(*[noise_ids, noise_augmentation_ids]))
    return sum([augmented_noise_length(noise_files[fi], noise_augmentations[ai]) for fi, ai in it])


def augmented_noise_length(noise_file: NoiseFile, noise_augmentation: Augmentation) -> int:
    from .augmentation import estimate_augmented_length_from_length

    return estimate_augmented_length_from_length(length=noise_file.samples,
                                                 tempo=noise_augmentation.tempo)



"""mkmanifest

usage: mkmanifest [-hvn] [--include GLOB] [-m METHOD] [-e ADAT] [-o OUTPUT] PATH ...

options:
    -h, --help
    -v, --verbose                   Be verbose: list all files found.
    -n, --dry-run                   Collect files, but exit without processing and writing manifest file.
    --include GLOB                  Search only files whose base name matches GLOB. [default: *.{wav,flac}].
    -m METHOD, --method METHOD      Method for getting the true speech text of the audio files. [default: librispeech].
    -e ADAT, --audio-env ADAT       Environment variable pointing to all audio data.
    -o OUTPUT, --output OUTPUT      Output file name. [default: asr_manifest.json].

Make a speech recognition (ASR) .json manifest file of all audio files under PATHS following the NVIDIA NeMo format.
An example of manifest entries:

{"audio_filepath": "<absolute_path_to>/1355-39947-0000.wav", "duration": 11.3, "text": "psychotherapy ..."}
{"audio_filepath": "<absolute_path_to>/1355-39947-0001.wav", "duration": 15.905, "text": "it is an ..."}

See the NVIDIA NeMo docs for more information:
https://docs.nvidia.com/deeplearning/nemo/user-guide/docs/en/stable/asr/datasets.html

Inputs:
    PATH        A relative path name or list of paths containing audio files. Each will be
                recursively searched for files matching the pattern GLOB.
    GLOB        Match the pattern GLOB using wildcard matching.
                Example: '*.{wav,flac}' matches all .wav and .flac files.
    METHOD      The method to use for fetching the true speech of the audio files.
                Supported methods:
                    - 'librispeech'
                    - 'vctk_noisy_speech' expects subdirs named like <name>_wav/ and <name>_txt/ with files in
                      each using same basename, but with .wav and .txt respectively.
    ADAT        Audio data environment variable. All found files will be expanded to their full, absolute path and
                then parts of the path that match the specified environment variable value will be replaced with
                the variable. This accommodates portability across platforms where the sound datasets may in
                different locations.
    OUTPUT      Name of output file. Default is asr_manifest.json.

Outputs the following to the current directory:
    <OUTPUT>
    mkmanifest.log

Example usage for LibriSpeech:
  sonusai mkmanifest -mlibrispeech -eADAT -oasr_manifest.json --include='*.flac' train-clean-100

"""
from sonusai import logger

VALID_METHOD = ['librispeech', 'vctk_noisy_speech']


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    verbose = args['--verbose']
    dry_run = args['--dry-run']
    include = args['--include']
    method = args['--method']
    audio_env = args['--audio-env']
    output = args['--output']
    paths = args['PATH']

    import json
    from functools import partial
    import time
    from os import environ
    from os.path import abspath
    from os.path import join
    from os.path import realpath

    from tqdm import tqdm

    from sonusai import SonusAIError
    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.utils import braced_iglob
    from sonusai.utils import pp_tqdm_imap
    from sonusai.utils import seconds_to_hms
    from sonusai.utils.asr_manifest_functions import PathInfo
    from sonusai.utils.asr_manifest_functions import collect_librispeech_transcripts
    from sonusai.utils.asr_manifest_functions import collect_vctk_noisy_speech_transcripts
    from sonusai.utils.asr_manifest_functions import get_librispeech_manifest_entry
    from sonusai.utils.asr_manifest_functions import get_vctk_noisy_speech_manifest_entry

    start_time = time.monotonic()

    create_file_handler('mkmanifest.log')
    update_console_handler(verbose)
    initial_log_messages('mkmanifest')

    if method not in VALID_METHOD:
        raise SonusAIError(f'Unknown method: {method}')

    audio_dir = None
    if audio_env is not None:
        audio_dir = realpath(environ[audio_env])
        if audio_dir is None:
            raise SonusAIError(f'Unknown environment variable: {audio_env}')

    if audio_env:
        for p in paths:
            if not realpath(abspath(p)).startswith(audio_dir):
                logger.warning(f'Specified directory, {p}, is not part of the provided audio environment: '
                               f'${audio_env}={audio_dir}')

    logger.info('')
    logger.info(f'Searching {len(paths)} provided director{"ies" if len(paths) > 1 else "y"}...')

    entries: list[PathInfo] = []
    for p in paths:
        location = join(realpath(abspath(p)), '**', include)
        logger.debug(f'Processing {location}')
        for file in braced_iglob(pathname=location, recursive=True):
            name = file
            if audio_env is not None:
                name = name.replace(audio_dir, f'${audio_env}')
            entries.append(PathInfo(abs_path=file, audio_filepath=name))
        logger.debug('')

    logger.info(f'Found {len(entries)} audio file{"s" if len(entries) != 1 else ""}')

    if dry_run:
        logger.info('')
        logger.info('Dry run')
        logger.info('')
        for entry in entries:
            logger.info(f' - {entry.audio_filepath}')
        return

    if method == 'librispeech':
        logger.info('Collecting LibriSpeech transcript data')
        transcript_data = collect_librispeech_transcripts(paths=paths)

        processing_func = partial(get_librispeech_manifest_entry, transcript_data=transcript_data)
        progress = tqdm(total=len(entries), desc='Creating LibriSpeech manifest data')
        results = pp_tqdm_imap(processing_func, entries, progress=progress)
        progress.close()

        with open(output, 'w') as f:
            for result in results:
                f.write(json.dumps(result) + '\n')

    if method == 'vctk_noisy_speech':
        logger.info('Collecting VCTK Noisy Speech transcript data')
        transcript_data = collect_vctk_noisy_speech_transcripts(paths=paths)

        processing_func = partial(get_vctk_noisy_speech_manifest_entry, transcript_data=transcript_data)
        progress = tqdm(total=len(entries), desc='Creating VCTK Noisy Speech manifest data')
        results = pp_tqdm_imap(processing_func, entries, progress=progress)
        progress.close()

        with open(output, 'w') as f:
            for result in results:
                f.write(json.dumps(result) + '\n')

    end_time = time.monotonic()
    logger.info('')
    logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')
    logger.info('')


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)



from functools import cached_property
from functools import lru_cache
from functools import partial
from sqlite3 import Connection
from sqlite3 import Cursor
from typing import Any
from typing import Optional

from sonusai.mixture.datatypes import AudioF
from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import AudiosF
from sonusai.mixture.datatypes import AudiosT
from sonusai.mixture.datatypes import ClassCount
from sonusai.mixture.datatypes import Feature
from sonusai.mixture.datatypes import FeatureGeneratorConfig
from sonusai.mixture.datatypes import FeatureGeneratorInfo
from sonusai.mixture.datatypes import GeneralizedIDs
from sonusai.mixture.datatypes import ImpulseResponseFiles
from sonusai.mixture.datatypes import Mixture
from sonusai.mixture.datatypes import Mixtures
from sonusai.mixture.datatypes import NoiseFile
from sonusai.mixture.datatypes import NoiseFiles
from sonusai.mixture.datatypes import Segsnr
from sonusai.mixture.datatypes import SpectralMask
from sonusai.mixture.datatypes import SpectralMasks
from sonusai.mixture.datatypes import TargetFile
from sonusai.mixture.datatypes import TargetFiles
from sonusai.mixture.datatypes import TransformConfig
from sonusai.mixture.datatypes import Truth
from sonusai.mixture.datatypes import UniversalSNR


def db_file(location: str, test: bool = False) -> str:
    from os.path import join

    if test:
        name = 'mixdb_test.db'
    else:
        name = 'mixdb.db'

    return join(location, name)


def db_connection(location: str, create: bool = False, readonly: bool = True, test: bool = False) -> Connection:
    import sqlite3
    from os import remove
    from os.path import exists

    from sonusai import SonusAIError

    name = db_file(location, test)
    if create and exists(name):
        remove(name)

    if not create and not exists(name):
        raise SonusAIError(f'Could not find mixture database in {location}')

    if not create and readonly:
        name += '?mode=ro'

    connection = sqlite3.connect('file:' + name, uri=True)
    # connection.set_trace_callback(print)
    return connection


class SQLiteContextManager:
    def __init__(self, location: str, test: bool = False) -> None:
        self.location = location
        self.test = test

    def __enter__(self) -> Cursor:
        self.con = db_connection(location=self.location, test=self.test)
        self.cur = self.con.cursor()
        return self.cur

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.con.close()


class MixtureDatabase:
    def __init__(self, location: str, test: bool = False) -> None:
        self.location = location
        self.db = partial(SQLiteContextManager, self.location, test)

    @cached_property
    def json(self) -> str:
        from .datatypes import MixtureDatabaseConfig

        config = MixtureDatabaseConfig(
            asr_manifest=self.asr_manifests,
            class_balancing=self.class_balancing,
            class_labels=self.class_labels,
            class_weights_threshold=self.class_weights_thresholds,
            feature=self.feature,
            impulse_response_files=self.impulse_response_files,
            mixtures=self.mixtures,
            noise_mix_mode=self.noise_mix_mode,
            noise_files=self.noise_files,
            num_classes=self.num_classes,
            spectral_masks=self.spectral_masks,
            target_files=self.target_files,
            truth_mutex=self.truth_mutex,
            truth_reduction_function=self.truth_reduction_function
        )
        return config.to_json(indent=2)

    def save(self) -> None:
        """Save the MixtureDatabase as a JSON file
        """
        from os.path import join

        json_name = join(self.location, 'mixdb.json')
        with open(file=json_name, mode='w') as file:
            file.write(self.json)

    def target_asr_data(self, t_id: int) -> str | None:
        """Get the ASR data for the given target ID

        :param t_id: Target ID
        :return: ASR text or None
        """
        from .tokenized_shell_vars import tokenized_expand

        name, _ = tokenized_expand(self.target_file(t_id).name)
        return self.asr_manifest_data.get(name, None)

    def mixture_asr_data(self, m_id: int) -> list[str | None]:
        """Get the ASR data for the given mixid

        :param m_id: Zero-based mixture ID
        :return: List of ASR text or None
        """
        return [self.target_asr_data(target.file_id) for target in self.mixture(m_id).targets]

    @cached_property
    def asr_manifest_data(self) -> dict[str, str]:
        """Get ASR data

        Each line of a manifest file should be in the following format:

        {"audio_filepath": "/path/to/audio.wav", "text": "the transcription of the utterance", "duration": 23.147}

        The audio_filepath field should provide an absolute path to the audio file corresponding to the utterance. The
        text field should contain the full transcript for the utterance, and the duration field should reflect the
        duration of the utterance in seconds.

        Each entry in the manifest (describing one audio file) should be bordered by '{' and '}' and must be contained
        on one line. The fields that describe the file should be separated by commas, and have the form
        "field_name": value, as shown above.

        Since the manifest specifies the path for each utterance, the audio files do not have to be located in the same
        directory as the manifest, or even in any specific directory structure.

        The manifest dictionary consists of key/value pairs where the keys are target file names and the values are ASR
        text.
        """
        import json

        from sonusai import SonusAIError
        from .tokenized_shell_vars import tokenized_expand

        expected_keys = ['audio_filepath', 'text', 'duration']

        def _error_preamble(e_name: str, e_line_num: int) -> str:
            return f'Invalid entry in ASR manifest {e_name} line {e_line_num}'

        asr_manifest_data: dict[str, str] = {}

        for name in self.asr_manifests:
            expanded_name, _ = tokenized_expand(name)
            with open(file=expanded_name, mode='r') as f:
                line_num = 1
                for line in f:
                    result = json.loads(line.strip())

                    for key in expected_keys:
                        if key not in result:
                            SonusAIError(f'{_error_preamble(name, line_num)}: missing field "{key}"')

                    for key in result.keys():
                        if key not in expected_keys:
                            SonusAIError(f'{_error_preamble(name, line_num)}: unknown field "{key}"')

                    key, _ = tokenized_expand(result['audio_filepath'])
                    value = result['text']

                    if key in asr_manifest_data:
                        SonusAIError(f'{_error_preamble(name, line_num)}: entry already exists')

                    asr_manifest_data[key] = value

                    line_num += 1

        return asr_manifest_data

    @cached_property
    def fg_config(self) -> FeatureGeneratorConfig:
        return FeatureGeneratorConfig(feature_mode=self.feature,
                                      num_classes=self.num_classes,
                                      truth_mutex=self.truth_mutex)

    @cached_property
    def fg_info(self) -> FeatureGeneratorInfo:
        from .helpers import get_feature_generator_info

        return get_feature_generator_info(self.fg_config)

    @cached_property
    def num_classes(self) -> int:
        with self.db() as c:
            return int(c.execute("SELECT top.num_classes from top").fetchone()[0])

    @cached_property
    def truth_mutex(self) -> bool:
        with self.db() as c:
            return bool(c.execute("SELECT top.truth_mutex from top").fetchone()[0])

    @cached_property
    def truth_reduction_function(self) -> str:
        with self.db() as c:
            return str(c.execute("SELECT top.truth_reduction_function from top").fetchone()[0])

    @cached_property
    def noise_mix_mode(self) -> str:
        with self.db() as c:
            return str(c.execute("SELECT top.noise_mix_mode from top").fetchone()[0])

    @cached_property
    def class_balancing(self) -> bool:
        with self.db() as c:
            return bool(c.execute("SELECT top.class_balancing from top").fetchone()[0])

    @cached_property
    def feature(self) -> str:
        with self.db() as c:
            return str(c.execute("SELECT top.feature from top").fetchone()[0])

    @cached_property
    def fg_decimation(self) -> int:
        return self.fg_info.decimation

    @cached_property
    def fg_stride(self) -> int:
        return self.fg_info.stride

    @cached_property
    def fg_step(self) -> int:
        return self.fg_info.step

    @cached_property
    def fg_num_bands(self) -> int:
        return self.fg_info.num_bands

    @cached_property
    def ft_config(self) -> TransformConfig:
        return self.fg_info.ft_config

    @cached_property
    def eft_config(self) -> TransformConfig:
        return self.fg_info.eft_config

    @cached_property
    def it_config(self) -> TransformConfig:
        return self.fg_info.it_config

    @cached_property
    def transform_frame_ms(self) -> float:
        from .constants import SAMPLE_RATE

        return float(self.ft_config.R) / float(SAMPLE_RATE / 1000)

    @cached_property
    def feature_ms(self) -> float:
        return self.transform_frame_ms * self.fg_decimation * self.fg_stride

    @cached_property
    def feature_samples(self) -> int:
        return self.ft_config.R * self.fg_decimation * self.fg_stride

    @cached_property
    def feature_step_ms(self) -> float:
        return self.transform_frame_ms * self.fg_decimation * self.fg_step

    @cached_property
    def feature_step_samples(self) -> int:
        return self.ft_config.R * self.fg_decimation * self.fg_step

    def total_samples(self, mixids: GeneralizedIDs = '*') -> int:
        return sum([self.mixture(m_id).samples for m_id in self.mixids_to_list(mixids)])

    def total_transform_frames(self, mixids: GeneralizedIDs = '*') -> int:
        return self.total_samples(mixids) // self.ft_config.R

    def total_feature_frames(self, mixids: GeneralizedIDs = '*') -> int:
        return self.total_samples(mixids) // self.feature_step_samples

    def mixture_transform_frames(self, samples: int) -> int:
        return samples // self.ft_config.R

    def mixture_feature_frames(self, samples: int) -> int:
        return samples // self.feature_step_samples

    def mixids_to_list(self, mixids: Optional[GeneralizedIDs] = None) -> list[int]:
        """Resolve generalized mixture IDs to a list of integers

        :param mixids: Generalized mixture IDs
        :return: List of mixture ID integers
        """
        from .helpers import generic_ids_to_list

        return generic_ids_to_list(self.num_mixtures, mixids)

    @cached_property
    def asr_manifests(self) -> list[str]:
        """Get ASR manifests from db

        :return: ASR manifests
        """
        with self.db() as c:
            return [str(item[0]) for item in c.execute("SELECT asr_manifest.manifest FROM asr_manifest").fetchall()]

    @cached_property
    def class_labels(self) -> list[str]:
        """Get class labels from db

        :return: Class labels
        """
        with self.db() as c:
            return [str(item[0]) for item in
                    c.execute("SELECT class_label.label FROM class_label ORDER BY class_label.id").fetchall()]

    @cached_property
    def class_weights_thresholds(self) -> list[float]:
        """Get class weights thresholds from db

        :return: Class weights thresholds
        """
        with self.db() as c:
            return [float(item[0]) for item in
                    c.execute("SELECT class_weights_threshold.threshold FROM class_weights_threshold").fetchall()]

    @cached_property
    def random_snrs(self) -> list[float]:
        """Get random snrs from db

        :return: Random SNRs
        """
        with self.db() as c:
            return list(set([float(item[0]) for item in
                             c.execute("SELECT mixture.snr FROM mixture WHERE mixture.random_snr == 1").fetchall()]))

    @cached_property
    def snrs(self) -> list[float]:
        """Get snrs from db

        :return: SNRs
        """
        with self.db() as c:
            return list(set([float(item[0]) for item in
                             c.execute("SELECT mixture.snr FROM mixture WHERE mixture.random_snr == 0").fetchall()]))

    @cached_property
    def all_snrs(self) -> list[UniversalSNR]:
        return sorted(list(set([UniversalSNR(is_random=False, value=snr) for snr in self.snrs] +
                               [UniversalSNR(is_random=True, value=snr) for snr in self.random_snrs])))

    @cached_property
    def spectral_masks(self) -> SpectralMasks:
        """Get spectral masks from db

        :return: Spectral masks
        """
        with self.db() as c:
            results = c.execute(
                "SELECT spectral_mask.f_max_width, f_num, t_max_width, t_num, t_max_percent FROM spectral_mask")
            return [SpectralMask(f_max_width=spectral_mask[0],
                                 f_num=spectral_mask[1],
                                 t_max_width=spectral_mask[2],
                                 t_num=spectral_mask[3],
                                 t_max_percent=spectral_mask[4]) for spectral_mask in results.fetchall()]

    def spectral_mask(self, sm_id: int) -> SpectralMask:
        """Get spectral mask with ID from db

        :param sm_id: Spectral mask ID
        :return: Spectral mask
        """
        return _spectral_mask(self.db, sm_id)

    @cached_property
    def target_files(self) -> TargetFiles:
        """Get target files from db

        :return: Target files
        """
        import json

        from .datatypes import TruthSetting
        from .datatypes import TruthSettings

        with self.db() as c:
            target_files: TargetFiles = []
            for target in c.execute("SELECT target_file.name, samples, level_type, id FROM target_file").fetchall():
                truth_settings: TruthSettings = []
                for ts in c.execute(
                        "SELECT truth_setting.setting " +
                        "FROM truth_setting, target_file_truth_setting " +
                        "WHERE ? = target_file_truth_setting.target_file_id " +
                        "AND truth_setting.id = target_file_truth_setting.truth_setting_id",
                        (target[3],)).fetchall():
                    entry = json.loads(ts[0])
                    truth_settings.append(TruthSetting(config=entry.get('config', None),
                                                       function=entry.get('function', None),
                                                       index=entry.get('index', None)))
                target_files.append(TargetFile(name=target[0],
                                               samples=target[1],
                                               level_type=target[2],
                                               truth_settings=truth_settings))
            return target_files

    @cached_property
    def target_file_ids(self) -> list[int]:
        """Get target file IDs from db

        :return: List of target file IDs
        """
        with self.db() as c:
            return [int(item[0]) for item in c.execute("SELECT target_file.id FROM target_file").fetchall()]

    def target_file(self, t_id: int) -> TargetFile:
        """Get target file with ID from db

        :param t_id: Target file ID
        :return: Target file
        """
        return _target_file(self.db, t_id)

    @cached_property
    def num_target_files(self) -> int:
        """Get number of target files from db

        :return: Number of target files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(target_file.id) FROM target_file").fetchone()[0])

    @cached_property
    def noise_files(self) -> NoiseFiles:
        """Get noise files from db

        :return: Noise files
        """
        with self.db() as c:
            return [NoiseFile(name=noise[0], samples=noise[1]) for noise in
                    c.execute("SELECT noise_file.name, samples FROM noise_file").fetchall()]

    @cached_property
    def noise_file_ids(self) -> list[int]:
        """Get noise file IDs from db

        :return: List of noise file IDs
        """
        with self.db() as c:
            return [int(item[0]) for item in c.execute("SELECT noise_file.id FROM noise_file").fetchall()]

    def noise_file(self, n_id: int) -> NoiseFile:
        """Get noise file with ID from db

        :param n_id: Noise file ID
        :return: Noise file
        """
        return _noise_file(self.db, n_id)

    @cached_property
    def num_noise_files(self) -> int:
        """Get number of noise files from db

        :return: Number of noise files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(noise_file.id) FROM noise_file").fetchone()[0])

    @cached_property
    def impulse_response_files(self) -> ImpulseResponseFiles:
        """Get impulse response files from db

        :return: Impulse response files
        """
        with self.db() as c:
            return [str(impulse_response[0]) for impulse_response in
                    c.execute("SELECT impulse_response_file.file FROM impulse_response_file").fetchall()]

    @cached_property
    def impulse_response_file_ids(self) -> list[int]:
        """Get impulse response file IDs from db

        :return: List of impulse response file IDs
        """
        with self.db() as c:
            return [int(item[0]) for item in
                    c.execute("SELECT impulse_response_file.id FROM impulse_response_file").fetchall()]

    def impulse_response_file(self, ir_id: int) -> str:
        """Get impulse response file with ID from db

        :param ir_id: Impulse response file ID
        :return: Noise
        """
        return _impulse_response_file(self.db, ir_id)

    @cached_property
    def num_impulse_response_files(self) -> int:
        """Get number of impulse response files from db

        :return: Number of impulse response files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(impulse_response_file.id) FROM impulse_response_file").fetchone()[0])

    @cached_property
    def mixtures(self) -> Mixtures:
        """Get mixtures from db

        :return: Mixtures
        """
        from .helpers import to_mixture
        from .helpers import to_target

        with self.db() as c:
            mixtures: Mixtures = []
            for mixture in c.execute(
                    "SELECT mixture.name, noise_file_id, noise_augmentation, noise_offset, noise_snr_gain, " +
                    "random_snr, snr, samples, spectral_mask_id, spectral_mask_seed, target_snr_gain, id " +
                    "FROM mixture").fetchall():
                targets = [to_target(target) for target in c.execute(
                    "SELECT target.file_id, augmentation, gain " +
                    "FROM target, mixture_target " +
                    "WHERE ? = mixture_target.mixture_id AND target.id = mixture_target.target_id",
                    (mixture[11],)).fetchall()]
                mixtures.append(to_mixture(mixture, targets))
            return mixtures

    @cached_property
    def mixture_ids(self) -> list[int]:
        """Get mixture IDs from db

        :return: List of zero-based mixture IDs
        """
        with self.db() as c:
            return [int(item[0]) - 1 for item in c.execute("SELECT mixture.id FROM mixture").fetchall()]

    def mixture(self, m_id: int) -> Mixture:
        """Get mixture record with ID from db

        :param m_id: Zero-based mixture ID
        :return: Mixture record
        """
        return _mixture(self.db, m_id)

    @cached_property
    def mixid_width(self) -> int:
        with self.db() as c:
            return int(c.execute("SELECT top.mixid_width from top").fetchone()[0])

    def location_filename(self, name: str) -> str:
        """Add the location to the given file name

        :param name: File name
        :return: Location added
        """
        from os.path import join

        return join(self.location, name)

    def mixture_filename(self, m_id: int) -> str:
        """Get the HDF5 file name for the give mixture ID

        :param m_id: Zero-based mixture ID
        :return: File name
        """
        return self.location_filename(self.mixture(m_id).name)

    @cached_property
    def num_mixtures(self) -> int:
        """Get number of mixtures from db

        :return: Number of mixtures
        """
        with self.db() as c:
            return int(c.execute("SELECT count(mixture.id) FROM mixture").fetchone()[0])

    def read_mixture_data(self, m_id: int, items: list[str] | str) -> Any:
        """Read mixture data from a mixture HDF5 file

        :param m_id: Zero-based mixture ID
        :param items: String(s) of dataset(s) to retrieve
        :return: Data (or tuple of data)
        """
        from .helpers import read_mixture_data

        return read_mixture_data(self.location_filename(self.mixture(m_id).name), items)

    def read_target_audio(self, t_id: int) -> AudioT:
        """Read target audio

        :param t_id: Target ID
        :return: Target audio
        """
        from .audio import read_audio

        return read_audio(self.target_file(t_id).name)

    def augmented_noise_audio(self, mixture: Mixture) -> AudioT:
        """Get augmented noise audio

        :param mixture: Mixture
        :return: Augmented noise audio
        """
        from .audio import read_audio
        from .audio import read_ir
        from .augmentation import apply_augmentation
        from .augmentation import apply_impulse_response

        noise = self.noise_file(mixture.noise.file_id)
        audio = read_audio(noise.name)
        audio = apply_augmentation(audio, mixture.noise.augmentation)
        if mixture.noise.augmentation.ir is not None:
            audio = apply_impulse_response(audio, read_ir(self.impulse_response_file(mixture.noise.augmentation.ir)))

        return audio

    def mixture_targets(self, m_id: int, force: bool = False) -> AudiosT:
        """Get the list of augmented target audio data (one per target in the mixup) for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: List of augmented target audio data (one per target in the mixup)
        """
        from sonusai import SonusAIError
        from .augmentation import apply_augmentation
        from .augmentation import apply_gain
        from .augmentation import pad_audio_to_length

        if not force:
            targets_audio = self.read_mixture_data(m_id, 'targets')
            if targets_audio is not None:
                return list(targets_audio)

        mixture = self.mixture(m_id)
        if mixture is None:
            raise SonusAIError(f'Could not find mixture for m_id: {m_id}')

        targets_audio = []
        for target in mixture.targets:
            target_audio = self.read_target_audio(target.file_id)
            target_audio = apply_augmentation(audio=target_audio,
                                              augmentation=target.augmentation,
                                              frame_length=self.feature_step_samples)
            target_audio = apply_gain(audio=target_audio, gain=mixture.target_snr_gain)
            target_audio = pad_audio_to_length(audio=target_audio, length=mixture.samples)
            targets_audio.append(target_audio)

        return targets_audio

    def mixture_targets_f(self,
                          m_id: int,
                          targets: Optional[AudiosT] = None,
                          force: bool = False) -> AudiosF:
        """Get the list of augmented target transform data (one per target in the mixup) for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: List of augmented target transform data (one per target in the mixup)
        """
        from .helpers import forward_transform

        if force or targets is None:
            targets = self.mixture_targets(m_id, force)

        return [forward_transform(target, self.ft_config) for target in targets]

    def mixture_target(self,
                       m_id: int,
                       targets: Optional[AudiosT] = None,
                       force: bool = False) -> AudioT:
        """Get the augmented target audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented target audio data
        """
        from .helpers import get_target

        if not force:
            target = self.read_mixture_data(m_id, 'target')
            if target is not None:
                return target

        if force or targets is None:
            targets = self.mixture_targets(m_id, force)

        return get_target(self, self.mixture(m_id), targets)

    def mixture_target_f(self,
                         m_id: int,
                         targets: Optional[AudiosT] = None,
                         target: Optional[AudioT] = None,
                         force: bool = False) -> AudioF:
        """Get the augmented target transform data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio for the given mixid
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented target transform data
        """
        from .helpers import forward_transform

        if force or target is None:
            target = self.mixture_target(m_id, targets, force)

        return forward_transform(target, self.ft_config)

    def mixture_noise(self,
                      m_id: int,
                      force: bool = False) -> AudioT:
        """Get the augmented noise audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented noise audio data
        """
        from .audio import get_next_noise
        from .augmentation import apply_gain

        if not force:
            noise = self.read_mixture_data(m_id, 'noise')
            if noise is not None:
                return noise

        mixture = self.mixture(m_id)
        noise = self.augmented_noise_audio(mixture)
        noise = get_next_noise(audio=noise, offset=mixture.noise.offset, length=mixture.samples)
        return apply_gain(audio=noise, gain=mixture.noise_snr_gain)

    def mixture_noise_f(self,
                        m_id: int,
                        noise: Optional[AudioT] = None,
                        force: bool = False) -> AudioF:
        """Get the augmented noise transform for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented noise transform data
        """
        from .helpers import forward_transform

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        return forward_transform(noise, self.ft_config)

    def mixture_mixture(self,
                        m_id: int,
                        targets: Optional[AudiosT] = None,
                        target: Optional[AudioT] = None,
                        noise: Optional[AudioT] = None,
                        force: bool = False) -> AudioT:
        """Get the mixture audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Mixture audio data
        """
        if not force:
            mixture = self.read_mixture_data(m_id, 'mixture')
            if mixture is not None:
                return mixture

        if force or target is None:
            target = self.mixture_target(m_id, targets, force)

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        return target + noise

    def mixture_mixture_f(self,
                          m_id: int,
                          targets: Optional[AudiosT] = None,
                          target: Optional[AudioT] = None,
                          noise: Optional[AudioT] = None,
                          mixture: Optional[AudioT] = None,
                          force: bool = False) -> AudioF:
        """Get the mixture transform for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param mixture: Mixture audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Mixture transform data
        """
        from .helpers import forward_transform

        if force or mixture is None:
            mixture = self.mixture_mixture(m_id, targets, target, noise, force)

        return forward_transform(mixture, self.ft_config)

    def mixture_truth_t(self,
                        m_id: int,
                        targets: Optional[AudiosT] = None,
                        noise: Optional[AudioT] = None,
                        mixture: Optional[AudioT] = None,
                        force: bool = False) -> Truth:
        """Get the truth_t data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixture ID
        :param noise: Augmented noise audio data for the given mixture ID
        :param mixture: Mixture audio data for the given mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: truth_t data
        """
        from .helpers import get_truth_t

        if not force:
            truth_t = self.read_mixture_data(m_id, 'truth_t')
            if truth_t is not None:
                return truth_t

        if force or targets is None:
            targets = self.mixture_targets(m_id, force)

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        if force or mixture is None:
            noise = self.mixture_mixture(m_id,
                                         targets=targets,
                                         noise=noise,
                                         force=force)

        return get_truth_t(self, self.mixture(m_id), targets, noise, mixture)

    def mixture_segsnr_t(self,
                         m_id: int,
                         targets: Optional[AudiosT] = None,
                         target: Optional[AudioT] = None,
                         noise: Optional[AudioT] = None,
                         force: bool = False) -> Segsnr:
        """Get the segsnr_t data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: segsnr_t data
        """
        from .helpers import get_segsnr_t

        if not force:
            segsnr_t = self.read_mixture_data(m_id, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t

        if force or target is None:
            target = self.mixture_target(m_id, targets, force)

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        return get_segsnr_t(self, self.mixture(m_id), target, noise)

    def mixture_segsnr(self,
                       m_id: int,
                       segsnr_t: Optional[Segsnr] = None,
                       targets: Optional[AudiosT] = None,
                       target: Optional[AudioT] = None,
                       noise: Optional[AudioT] = None,
                       force: bool = False) -> Segsnr:
        """Get the segsnr data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param segsnr_t: segsnr_t data
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: segsnr data
        """
        if not force:
            segsnr = self.read_mixture_data(m_id, 'segsnr')
            if segsnr is not None:
                return segsnr

            segsnr_t = self.read_mixture_data(m_id, 'segsnr_t')
            if segsnr_t is not None:
                return segsnr_t[0::self.ft_config.R]

        if force or segsnr_t is None:
            segsnr_t = self.mixture_segsnr_t(m_id, targets, target, noise, force)

        return segsnr_t[0::self.ft_config.R]

    def mixture_ft(self,
                   m_id: int,
                   targets: Optional[AudiosT] = None,
                   target: Optional[AudioT] = None,
                   noise: Optional[AudioT] = None,
                   mixture_f: Optional[AudioF] = None,
                   mixture: Optional[AudioT] = None,
                   truth_t: Optional[Truth] = None,
                   force: bool = False) -> tuple[Feature, Truth]:
        """Get the feature and truth_f data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param mixture_f: Mixture transform data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Tuple of (feature, truth_f) data
        """
        from dataclasses import asdict

        import numpy as np
        from pyaaware import FeatureGenerator

        from .spectral_mask import apply_spectral_mask
        from .truth import truth_reduction

        if not force:
            feature, truth_f = self.read_mixture_data(m_id, ['feature', 'truth_f'])
            if feature is not None and truth_f is not None:
                return feature, truth_f

        if force or mixture_f is None:
            mixture_f = self.mixture_mixture_f(m_id=m_id,
                                               targets=targets,
                                               target=target,
                                               noise=noise,
                                               mixture=mixture,
                                               force=force)

        if force or truth_t is None:
            truth_t = self.mixture_truth_t(m_id=m_id, targets=targets, noise=noise, force=force)

        m = self.mixture(m_id)
        transform_frames = self.mixture_transform_frames(m.samples)
        feature_frames = self.mixture_feature_frames(m.samples)

        if truth_t is None:
            truth_t = np.zeros((m.samples, self.num_classes), dtype=np.float32)

        feature = np.empty((feature_frames, self.fg_stride, self.fg_num_bands), dtype=np.float32)
        truth_f = np.empty((feature_frames, self.num_classes), dtype=np.complex64)

        fg = FeatureGenerator(**asdict(self.fg_config))
        feature_frame = 0
        for transform_frame in range(transform_frames):
            indices = slice(transform_frame * self.ft_config.R, (transform_frame + 1) * self.ft_config.R)
            fg.execute(mixture_f[transform_frame],
                       truth_reduction(truth_t[indices], self.truth_reduction_function))

            if fg.eof():
                feature[feature_frame] = fg.feature()
                truth_f[feature_frame] = fg.truth()
                feature_frame += 1

        if m.spectral_mask_id is not None:
            feature = apply_spectral_mask(feature=feature,
                                          spectral_mask=self.spectral_mask(int(m.spectral_mask_id)),
                                          seed=m.spectral_mask_seed)

        if np.isreal(truth_f).all():
            return feature, truth_f.real

        return feature, truth_f

    def mixture_feature(self,
                        m_id: int,
                        targets: Optional[AudiosT] = None,
                        noise: Optional[AudioT] = None,
                        mixture: Optional[AudioT] = None,
                        truth_t: Optional[Truth] = None,
                        force: bool = False) -> Feature:
        """Get the feature data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param noise: Augmented noise audio data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Feature data
        """
        feature, _ = self.mixture_ft(m_id=m_id,
                                     targets=targets,
                                     noise=noise,
                                     mixture=mixture,
                                     truth_t=truth_t,
                                     force=force)
        return feature

    def mixture_truth_f(self,
                        m_id: int,
                        targets: Optional[AudiosT] = None,
                        noise: Optional[AudioT] = None,
                        mixture: Optional[AudioT] = None,
                        truth_t: Optional[Truth] = None,
                        force: bool = False) -> Truth:
        """Get the truth_f data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param noise: Augmented noise audio data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: truth_f data
        """
        _, truth_f = self.mixture_ft(m_id=m_id,
                                     targets=targets,
                                     noise=noise,
                                     mixture=mixture,
                                     truth_t=truth_t,
                                     force=force)
        return truth_f

    def mixture_class_count(self,
                            m_id: int,
                            targets: Optional[AudiosT] = None,
                            noise: Optional[AudioT] = None,
                            truth_t: Optional[Truth] = None) -> ClassCount:
        """Compute the number of samples for which each truth index is active for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio (one per target in the mixup)
        :param noise: Augmented noise audio
        :param truth_t: truth_t
        :return: List of class counts
        """
        import numpy as np

        if truth_t is None:
            truth_t = self.mixture_truth_t(m_id, targets, noise)

        class_count = [0] * self.num_classes
        num_classes = self.num_classes
        if self.truth_mutex:
            num_classes -= 1
        for cl in range(num_classes):
            class_count[cl] = int(np.sum(truth_t[:, cl] >= self.class_weights_thresholds[cl]))

        return class_count


@lru_cache
def _spectral_mask(db: partial, sm_id: int) -> SpectralMask:
    """Get spectral mask with ID from db

    :param db: Database context
    :param sm_id: Spectral mask ID
    :return: Spectral mask
    """
    with db() as c:
        spectral_mask = c.execute(
            "SELECT spectral_mask.f_max_width, f_num, t_max_width, t_num, t_max_percent " +
            "FROM spectral_mask " +
            "WHERE ? = spectral_mask.id",
            (sm_id,)).fetchone()
        return SpectralMask(f_max_width=spectral_mask[0],
                            f_num=spectral_mask[1],
                            t_max_width=spectral_mask[2],
                            t_num=spectral_mask[3],
                            t_max_percent=spectral_mask[4])


@lru_cache
def _target_file(db: partial, t_id: int) -> TargetFile:
    """Get target file with ID from db

    :param db: Database context
    :param t_id: Target file ID
    :return: Target file
    """
    import json

    from .datatypes import TruthSetting
    from .datatypes import TruthSettings

    with db() as c:
        target = c.execute("SELECT target_file.name, samples, level_type FROM target_file WHERE ? = target_file.id",
                           (t_id,)).fetchone()

        truth_settings: TruthSettings = []
        for ts in c.execute(
                "SELECT truth_setting.setting " +
                "FROM truth_setting, target_file_truth_setting " +
                "WHERE ? = target_file_truth_setting.target_file_id " +
                "AND truth_setting.id = target_file_truth_setting.truth_setting_id",
                (t_id,)).fetchall():
            entry = json.loads(ts[0])
            truth_settings.append(TruthSetting(config=entry.get('config', None),
                                               function=entry.get('function', None),
                                               index=entry.get('index', None)))
        return TargetFile(name=target[0],
                          samples=target[1],
                          level_type=target[2],
                          truth_settings=truth_settings)


@lru_cache
def _noise_file(db: partial, n_id: int) -> NoiseFile:
    """Get noise file with ID from db

    :param db: Database context
    :param n_id: Noise file ID
    :return: Noise file
    """
    with db() as c:
        noise = c.execute("SELECT noise_file.name, samples FROM noise_file WHERE ? = noise_file.id",
                          (n_id,)).fetchone()
        return NoiseFile(name=noise[0], samples=noise[1])


@lru_cache
def _impulse_response_file(db: partial, ir_id: int) -> str:
    """Get impulse response file with ID from db

    :param db: Database context
    :param ir_id: Impulse response file ID
    :return: Noise
    """
    with db() as c:
        return str(c.execute(
            "SELECT impulse_response_file.file FROM impulse_response_file WHERE ? = impulse_response_file.id",
            (ir_id + 1,)).fetchone()[0])


@lru_cache
def _mixture(db: partial, m_id: int) -> Mixture:
    """Get mixture record with ID from db

    :param db: Database context
    :param m_id: Zero-based mixture ID
    :return: Mixture record
    """
    from .helpers import to_mixture
    from .helpers import to_target

    with db() as c:
        mixture = c.execute(
            "SELECT mixture.name, noise_file_id, noise_augmentation, noise_offset, noise_snr_gain, " +
            "random_snr, snr, samples, spectral_mask_id, spectral_mask_seed, target_snr_gain, id " +
            "FROM mixture " +
            "WHERE ? = mixture.id",
            (m_id + 1,)).fetchone()

        targets = [to_target(target) for target in c.execute(
            "SELECT target.file_id, augmentation, gain " +
            "FROM target, mixture_target " +
            "WHERE ? = mixture_target.mixture_id AND target.id = mixture_target.target_id",
            (mixture[11],)).fetchall()]

        return to_mixture(mixture, targets)



from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import AudiosT
from sonusai.mixture.datatypes import Augmentation
from sonusai.mixture.datatypes import AugmentationRules
from sonusai.mixture.datatypes import AugmentedTargets
from sonusai.mixture.datatypes import GenMixData
from sonusai.mixture.datatypes import ImpulseResponseFiles
from sonusai.mixture.datatypes import Mixture
from sonusai.mixture.datatypes import Mixtures
from sonusai.mixture.datatypes import NoiseFiles
from sonusai.mixture.datatypes import SpectralMasks
from sonusai.mixture.datatypes import TargetFiles
from sonusai.mixture.datatypes import Targets
from sonusai.mixture.datatypes import UniversalSNRGenerator
from sonusai.mixture.mixdb import MixtureDatabase


def config_file(location: str) -> str:
    from os.path import join

    return join(location, 'config.yml')


def initialize_db(location: str, test: bool = False) -> None:
    from .mixdb import db_connection

    con = db_connection(location=location, create=True, test=test)

    con.execute("""
    CREATE TABLE truth_setting(
    id INTEGER PRIMARY KEY NOT NULL,
    setting TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE target_file (
    id INTEGER PRIMARY KEY NOT NULL,
    name TEXT NOT NULL,
    samples INTEGER NOT NULL,
    level_type TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE noise_file (
    id INTEGER PRIMARY KEY NOT NULL,
    name TEXT NOT NULL,
    samples INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE top (
    id INTEGER PRIMARY KEY NOT NULL,
    version INTEGER NOT NULL,
    class_balancing BOOLEAN NOT NULL,
    feature TEXT NOT NULL,
    noise_mix_mode TEXT NOT NULL,
    num_classes INTEGER NOT NULL,
    seed INTEGER NOT NULL,
    truth_mutex BOOLEAN NOT NULL,
    truth_reduction_function TEXT NOT NULL,
    mixid_width INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE asr_manifest (
    id INTEGER PRIMARY KEY NOT NULL,
    manifest TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE class_label (
    id INTEGER PRIMARY KEY NOT NULL,
    label TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE class_weights_threshold (
    id INTEGER PRIMARY KEY NOT NULL,
    threshold FLOAT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE impulse_response_file (
    id INTEGER PRIMARY KEY NOT NULL,
    file TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE spectral_mask (
    id INTEGER PRIMARY KEY NOT NULL,
    f_max_width INTEGER NOT NULL,
    f_num INTEGER NOT NULL,
    t_max_width INTEGER NOT NULL,
    t_num INTEGER NOT NULL,
    t_max_percent INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE target_file_truth_setting (
    target_file_id INTEGER,
    truth_setting_id INTEGER,
    FOREIGN KEY(target_file_id) REFERENCES target_file (id),
    FOREIGN KEY(truth_setting_id) REFERENCES truth_setting (id))
    """)

    con.execute("""
    CREATE TABLE target (
    id INTEGER PRIMARY KEY NOT NULL,
    file_id INTEGER NOT NULL,
    augmentation TEXT NOT NULL,
    gain FLOAT,
    FOREIGN KEY(file_id) REFERENCES target_file (id))
    """)

    con.execute("""
    CREATE TABLE mixture (
    id INTEGER PRIMARY KEY NOT NULL,
    name VARCHAR NOT NULL,
    noise_file_id INTEGER NOT NULL,
    noise_augmentation TEXT NOT NULL,
    noise_offset INTEGER NOT NULL,
    noise_snr_gain FLOAT,
    random_snr BOOLEAN NOT NULL,
    snr FLOAT NOT NULL,
    samples INTEGER NOT NULL,
    spectral_mask_id INTEGER NOT NULL,
    spectral_mask_seed INTEGER NOT NULL,
    target_snr_gain FLOAT,
    FOREIGN KEY(noise_file_id) REFERENCES noise_file (id),
    FOREIGN KEY(spectral_mask_id) REFERENCES spectral_mask (id))
    """)

    con.execute("""
    CREATE TABLE mixture_target (
    mixture_id INTEGER,
    target_id INTEGER,
    FOREIGN KEY(mixture_id) REFERENCES mixture (id),
    FOREIGN KEY(target_id) REFERENCES target (id))
    """)

    con.commit()
    con.close()


def populate_top_table(location: str, config: dict, test: bool = False) -> None:
    """Populate top table
    """
    from sonusai import SonusAIError
    from .mixdb import db_connection

    if config['truth_mode'] not in ['normal', 'mutex']:
        raise SonusAIError(f'invalid truth_mode: {config["truth_mode"]}')
    truth_mutex = config['truth_mode'] == 'mutex'

    con = db_connection(location=location, readonly=False, test=test)
    con.execute("""
    INSERT INTO top (version, class_balancing, feature, noise_mix_mode, num_classes,
    seed, truth_mutex, truth_reduction_function, mixid_width)
    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
    """, (
        1,
        config['class_balancing'],
        config['feature'],
        config['noise_mix_mode'],
        config['num_classes'],
        config['seed'],
        truth_mutex,
        config['truth_reduction_function'],
        0))
    con.commit()
    con.close()


def populate_asr_manifest_table(location: str, config: dict, test: bool = False) -> None:
    """Populate asr_manifest table
    """
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany("INSERT INTO asr_manifest (manifest) VALUES (?)",
                    [(item,) for item in config['asr_manifest']])
    con.commit()
    con.close()


def populate_class_label_table(location: str, config: dict, test: bool = False) -> None:
    """Populate class_label table
    """
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany("INSERT INTO class_label (label) VALUES (?)",
                    [(item,) for item in config['class_labels']])
    con.commit()
    con.close()


def populate_class_weights_threshold_table(location: str, config: dict, test: bool = False) -> None:
    """Populate class_weights_threshold table
    """
    from sonusai import SonusAIError
    from .mixdb import db_connection

    class_weights_threshold = config['class_weights_threshold']
    num_classes = config['num_classes']

    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold]

    if len(class_weights_threshold) == 1:
        class_weights_threshold = [class_weights_threshold[0]] * num_classes

    if len(class_weights_threshold) != num_classes:
        raise SonusAIError(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany("INSERT INTO class_weights_threshold (threshold) VALUES (?)",
                    [(item,) for item in class_weights_threshold])
    con.commit()
    con.close()


def populate_spectral_mask_table(location: str, config: dict, test: bool = False) -> None:
    """Populate spectral_mask table
    """
    from .config import get_spectral_masks
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany("""
    INSERT INTO spectral_mask (f_max_width, f_num, t_max_width, t_num, t_max_percent) VALUES (?, ?, ?, ?, ?)
    """, [(item.f_max_width,
           item.f_num,
           item.t_max_width,
           item.t_num,
           item.t_max_percent) for item in get_spectral_masks(config)]
                    )
    con.commit()
    con.close()


def populate_target_file_table(location: str, target_files: TargetFiles, test: bool = False) -> None:
    """Populate target file table
    """
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)

    # Populate truth_setting table
    truth_settings: list[str] = []
    for truth_setting in [truth_setting for target_file in target_files
                          for truth_setting in target_file.truth_settings]:
        ts = truth_setting.to_json()
        if ts not in truth_settings:
            truth_settings.append(ts)
    con.executemany("INSERT INTO truth_setting (setting) VALUES (?)",
                    [(item,) for item in truth_settings])

    # Populate target_file table
    cur = con.cursor()
    for target_file in target_files:
        truth_setting_ids: list[int] = []
        for truth_setting in target_file.truth_settings:
            cur.execute("SELECT truth_setting.id FROM truth_setting WHERE ? = truth_setting.setting",
                        (truth_setting.to_json(),))
            truth_setting_ids.append(cur.fetchone()[0])

        cur.execute("INSERT INTO target_file (name, samples, level_type) VALUES (?, ?, ?)",
                    (target_file.name, target_file.samples, target_file.level_type))
        target_file_id = cur.lastrowid
        for truth_setting_id in truth_setting_ids:
            cur.execute("INSERT INTO target_file_truth_setting (target_file_id, truth_setting_id) VALUES (?, ?)",
                        (target_file_id, truth_setting_id))

    con.commit()
    con.close()


def populate_noise_file_table(location: str, noise_files: NoiseFiles, test: bool = False) -> None:
    """Populate noise file table
    """
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany("INSERT INTO noise_file (name, samples) VALUES (?, ?)",
                    [(noise_file.name, noise_file.samples) for noise_file in noise_files])
    con.commit()
    con.close()


def populate_impulse_response_file_table(location: str, impulse_response_files: ImpulseResponseFiles,
                                         test: bool = False) -> None:
    """Populate impulse response file table
    """
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany("INSERT INTO impulse_response_file (file) VALUES (?)",
                    [(impulse_response_file,) for impulse_response_file in impulse_response_files])
    con.commit()
    con.close()


def update_mixid_width(location: str, num_mixtures: int, test: bool = False) -> None:
    """Update the mixid width
    """
    from sonusai.utils import max_text_width
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.execute("UPDATE top SET mixid_width=? WHERE top.id = ?", (max_text_width(num_mixtures), 1))
    con.commit()
    con.close()


def populate_mixture_table(location: str, mixtures: Mixtures, test: bool = False) -> None:
    """Populate mixture table
    """
    from .helpers import from_mixture
    from .helpers import from_target
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)

    # Populate target table
    targets: list[tuple[int, str, float]] = []
    for mixture in mixtures:
        for target in mixture.targets:
            entry = from_target(target)
            if entry not in targets:
                targets.append(entry)

    con.executemany("INSERT INTO target (file_id, augmentation, gain) VALUES (?, ?, ?)", targets)

    # Populate mixture table
    cur = con.cursor()
    for mixture in mixtures:
        cur.execute("""
        INSERT INTO mixture (name, noise_file_id, noise_augmentation, noise_offset, noise_snr_gain, random_snr,
        snr, samples, spectral_mask_id, spectral_mask_seed, target_snr_gain)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
        """, from_mixture(mixture))

        mixture_id = cur.lastrowid
        for target in mixture.targets:
            target_id = con.execute("""
            SELECT target.id
            FROM target
            WHERE ? = target.file_id AND ? = target.augmentation AND ? = target.gain
            """, from_target(target)).fetchone()[0]
            con.execute("INSERT INTO mixture_target (mixture_id, target_id) VALUES (?, ?)",
                        (mixture_id, target_id))

    con.commit()
    con.close()


def update_mixture(mixdb: MixtureDatabase,
                   mixture: Mixture,
                   with_data: bool = False) -> tuple[Mixture, GenMixData]:
    """Update mixture record with name and gains

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param with_data: Return audio data
    :return: Generated audio data (if requested)
    """
    from .audio import get_next_noise
    from .augmentation import apply_gain
    from .helpers import get_target
    from .datatypes import GenMixData

    mixture, targets_audio = _initialize_targets_audio(mixdb, mixture)

    noise_audio = _augmented_noise_audio(mixdb, mixture)
    noise_audio = get_next_noise(audio=noise_audio, offset=mixture.noise.offset, length=mixture.samples)

    # Apply IR and sum targets audio before initializing the mixture SNR gains
    target_audio = get_target(mixdb, mixture, targets_audio)

    mixture = _initialize_mixture_gains(mixdb=mixdb,
                                        mixture=mixture,
                                        target_audio=target_audio,
                                        noise_audio=noise_audio)

    mixture.name = f'{int(mixture.name):0{mixdb.mixid_width}}.h5'

    if not with_data:
        return mixture, GenMixData()

    # Apply SNR gains
    targets_audio = [apply_gain(audio=target_audio, gain=mixture.target_snr_gain) for target_audio in targets_audio]
    noise_audio = apply_gain(audio=noise_audio, gain=mixture.noise_snr_gain)

    # Apply IR and sum targets audio after applying the mixture SNR gains
    target_audio = get_target(mixdb, mixture, targets_audio)
    mixture_audio = target_audio + noise_audio

    return mixture, GenMixData(mixture=mixture_audio,
                               targets=targets_audio,
                               target=target_audio,
                               noise=noise_audio)


def _augmented_noise_audio(mixdb: MixtureDatabase, mixture: Mixture) -> AudioT:
    from .audio import read_audio
    from .audio import read_ir
    from .augmentation import apply_augmentation
    from .augmentation import apply_impulse_response

    noise = mixdb.noise_file(mixture.noise.file_id)
    noise_augmentation = mixture.noise.augmentation

    audio = read_audio(noise.name)
    audio = apply_augmentation(audio, noise_augmentation)
    if noise_augmentation.ir is not None:
        audio = apply_impulse_response(audio, read_ir(mixdb.impulse_response_file(noise_augmentation.ir)))

    return audio


def _initialize_targets_audio(mixdb: MixtureDatabase, mixture: Mixture) -> tuple[Mixture, AudiosT]:
    from .augmentation import apply_augmentation
    from .augmentation import pad_audio_to_length

    targets_audio = []
    for target in mixture.targets:
        target_audio = mixdb.read_target_audio(target.file_id)
        targets_audio.append(apply_augmentation(audio=target_audio,
                                                augmentation=target.augmentation,
                                                frame_length=mixdb.feature_step_samples))

        # target_gain is used to back out the gain augmentation in order to return the target audio
        # to its normalized level when calculating truth (if needed).
        if target.augmentation.gain is not None:
            target.gain = round(10 ** (target.augmentation.gain / 20), ndigits=5)
        else:
            target.gain = 1

    mixture.samples = max([len(item) for item in targets_audio])

    for idx in range(len(targets_audio)):
        targets_audio[idx] = pad_audio_to_length(audio=targets_audio[idx], length=mixture.samples)

    return mixture, targets_audio


def _initialize_mixture_gains(mixdb: MixtureDatabase,
                              mixture: Mixture,
                              target_audio: AudioT,
                              noise_audio: AudioT) -> Mixture:
    import numpy as np

    from sonusai import SonusAIError
    from sonusai.utils import asl_p56
    from sonusai.utils import db_to_linear

    if mixture.snr < -96:
        # Special case for zeroing out target data
        mixture.target_snr_gain = 0
        mixture.noise_snr_gain = 1
        # Setting target_gain to zero will cause the truth to be all zeros.
        for target in mixture.targets:
            target.gain = 0
    elif mixture.snr > 96:
        # Special case for zeroing out noise data
        mixture.target_snr_gain = 1
        mixture.noise_snr_gain = 0
    else:
        target_level_types = [target_file.level_type for target_file in
                              [mixdb.target_file(target.file_id) for target in mixture.targets]]
        if not all(level_type == target_level_types[0] for level_type in target_level_types):
            raise SonusAIError(f'Not all target_level_types in mixup are the same')

        level_type = target_level_types[0]
        match level_type:
            case 'default':
                target_energy = np.mean(np.square(target_audio))
            case 'speech':
                target_energy = asl_p56(target_audio)
            case _:
                raise SonusAIError(f'Unknown level_type: {level_type}')

        noise_energy = np.mean(np.square(noise_audio))
        if noise_energy == 0:
            noise_gain = 1
        else:
            noise_gain = np.sqrt(target_energy / noise_energy) / db_to_linear(mixture.snr)

        # Check for noise_gain > 1 to avoid clipping
        if noise_gain > 1:
            mixture.target_snr_gain = 1 / noise_gain
            mixture.noise_snr_gain = 1
        else:
            mixture.target_snr_gain = 1
            mixture.noise_snr_gain = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = target_audio * mixture.target_snr_gain
    gain_adjusted_noise_audio = noise_audio * mixture.noise_snr_gain
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio
    max_abs_audio = max(abs(mixture_audio))
    clip_level = db_to_linear(-0.25)
    if max_abs_audio > clip_level:
        # Clipping occurred; lower gains to bring audio within +/-1
        gain_adjustment = clip_level / max_abs_audio
        mixture.target_snr_gain *= gain_adjustment
        mixture.noise_snr_gain *= gain_adjustment

    mixture.target_snr_gain = round(mixture.target_snr_gain, ndigits=5)
    mixture.noise_snr_gain = round(mixture.noise_snr_gain, ndigits=5)
    return mixture


def generate_mixtures(noise_mix_mode: str,
                      augmented_targets: AugmentedTargets,
                      target_files: TargetFiles,
                      target_augmentations: AugmentationRules,
                      noise_files: NoiseFiles,
                      noise_augmentations: AugmentationRules,
                      spectral_masks: SpectralMasks,
                      all_snrs: list[UniversalSNRGenerator],
                      mixups: list[int],
                      num_classes: int,
                      truth_mutex: bool,
                      feature_step_samples: int,
                      num_ir: int) -> tuple[int, int, Mixtures]:
    """Generate mixtures

    :param noise_mix_mode: Noise mix mode
    :param augmented_targets: List of augmented targets
    :param target_files: List of target files
    :param target_augmentations: List of target augmentations
    :param noise_files: List of noise files
    :param noise_augmentations: List of noise augmentations
    :param spectral_masks: List of spectral masks
    :param all_snrs: List of all SNRs
    :param mixups: List of mixup values
    :param num_classes: Number of classes
    :param truth_mutex: Truth mutex mode
    :param feature_step_samples: Number of samples in a feature step
    :param num_ir: Number of impulse response files
    :return: (Number of noise files used, number of noise samples used, list of mixture records)
    """
    from sonusai import SonusAIError

    if noise_mix_mode == 'exhaustive':
        return _exhaustive_noise_mix(augmented_targets=augmented_targets,
                                     target_files=target_files,
                                     target_augmentations=target_augmentations,
                                     noise_files=noise_files,
                                     noise_augmentations=noise_augmentations,
                                     spectral_masks=spectral_masks,
                                     all_snrs=all_snrs,
                                     mixups=mixups,
                                     num_classes=num_classes,
                                     truth_mutex=truth_mutex,
                                     feature_step_samples=feature_step_samples,
                                     num_ir=num_ir)

    if noise_mix_mode == 'non-exhaustive':
        return _non_exhaustive_noise_mix(augmented_targets=augmented_targets,
                                         target_files=target_files,
                                         target_augmentations=target_augmentations,
                                         noise_files=noise_files,
                                         noise_augmentations=noise_augmentations,
                                         spectral_masks=spectral_masks,
                                         all_snrs=all_snrs,
                                         mixups=mixups,
                                         num_classes=num_classes,
                                         truth_mutex=truth_mutex,
                                         feature_step_samples=feature_step_samples,
                                         num_ir=num_ir)

    if noise_mix_mode == 'non-combinatorial':
        return _non_combinatorial_noise_mix(augmented_targets=augmented_targets,
                                            target_files=target_files,
                                            target_augmentations=target_augmentations,
                                            noise_files=noise_files,
                                            noise_augmentations=noise_augmentations,
                                            spectral_masks=spectral_masks,
                                            all_snrs=all_snrs,
                                            mixups=mixups,
                                            num_classes=num_classes,
                                            truth_mutex=truth_mutex,
                                            feature_step_samples=feature_step_samples,
                                            num_ir=num_ir)

    raise SonusAIError(f'invalid noise_mix_mode: {noise_mix_mode}')


def _exhaustive_noise_mix(augmented_targets: AugmentedTargets,
                          target_files: TargetFiles,
                          target_augmentations: AugmentationRules,
                          noise_files: NoiseFiles,
                          noise_augmentations: AugmentationRules,
                          spectral_masks: SpectralMasks,
                          all_snrs: list[UniversalSNRGenerator],
                          mixups: list[int],
                          num_classes: int,
                          truth_mutex: bool,
                          feature_step_samples: int,
                          num_ir: int) -> tuple[int, int, Mixtures]:
    """ Use every noise/augmentation with every target/augmentation
    """
    from random import randint

    import numpy as np

    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length
    from .datatypes import Mixture
    from .datatypes import Mixtures
    from .datatypes import Noise
    from .datatypes import UniversalSNR
    from .targets import get_augmented_target_ids_for_mixup

    mixtures: Mixtures = []
    m_id = 0
    used_noise_files = len(noise_files) * len(noise_augmentations)
    used_noise_samples = 0

    augmented_target_ids_for_mixups = [get_augmented_target_ids_for_mixup(augmented_targets=augmented_targets,
                                                                          targets=target_files,
                                                                          target_augmentations=target_augmentations,
                                                                          mixup=mixup,
                                                                          num_classes=num_classes,
                                                                          truth_mutex=truth_mutex) for mixup in mixups]
    for noise_file_id in range(len(noise_files)):
        for noise_augmentation_rule in noise_augmentations:
            noise_augmentation = augmentation_from_rule(noise_augmentation_rule, num_ir)
            noise_offset = 0
            noise_length = estimate_augmented_length_from_length(
                length=noise_files[noise_file_id].samples,
                tempo=noise_augmentation.tempo)

            for augmented_target_ids_for_mixup in augmented_target_ids_for_mixups:
                for augmented_target_ids in augmented_target_ids_for_mixup:
                    targets, target_length = _get_target_info(augmented_target_ids=augmented_target_ids,
                                                              augmented_targets=augmented_targets,
                                                              target_files=target_files,
                                                              target_augmentations=target_augmentations,
                                                              feature_step_samples=feature_step_samples,
                                                              num_ir=num_ir)

                    for spectral_mask_id in range(len(spectral_masks)):
                        for snr in all_snrs:
                            mixtures.append(Mixture(
                                targets=targets,
                                name=str(m_id),
                                noise=Noise(file_id=noise_file_id + 1,
                                            augmentation=noise_augmentation,
                                            offset=noise_offset),
                                samples=target_length,
                                snr=UniversalSNR(value=snr.value,
                                                 is_random=snr.is_random),
                                spectral_mask_id=spectral_mask_id + 1,
                                spectral_mask_seed=randint(0, np.iinfo('i').max)))
                            m_id += 1

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length

    return used_noise_files, used_noise_samples, mixtures


def _non_exhaustive_noise_mix(augmented_targets: AugmentedTargets,
                              target_files: TargetFiles,
                              target_augmentations: AugmentationRules,
                              noise_files: NoiseFiles,
                              noise_augmentations: AugmentationRules,
                              spectral_masks: SpectralMasks,
                              all_snrs: list[UniversalSNRGenerator],
                              mixups: list[int],
                              num_classes: int,
                              truth_mutex: bool,
                              feature_step_samples: int,
                              num_ir: int) -> tuple[int, int, Mixtures]:
    """ Cycle through every target/augmentation without necessarily using all noise/augmentation combinations
    (reduced data set).
    """
    from random import randint

    import numpy as np

    from .datatypes import Mixture
    from .datatypes import Mixtures
    from .datatypes import Noise
    from .datatypes import UniversalSNR
    from .targets import get_augmented_target_ids_for_mixup

    mixtures: Mixtures = []
    m_id = 0
    used_noise_files = set()
    used_noise_samples = 0
    noise_file_id = None
    noise_augmentation_id = None
    noise_offset = None

    augmented_target_indices_for_mixups = [get_augmented_target_ids_for_mixup(
        augmented_targets=augmented_targets,
        targets=target_files,
        target_augmentations=target_augmentations,
        mixup=mixup,
        num_classes=num_classes,
        truth_mutex=truth_mutex) for mixup in mixups]
    for mixup in augmented_target_indices_for_mixups:
        for augmented_target_indices in mixup:
            targets, target_length = _get_target_info(augmented_target_ids=augmented_target_indices,
                                                      augmented_targets=augmented_targets,
                                                      target_files=target_files,
                                                      target_augmentations=target_augmentations,
                                                      feature_step_samples=feature_step_samples,
                                                      num_ir=num_ir)

            for spectral_mask_id in range(len(spectral_masks)):
                for snr in all_snrs:
                    (noise_file_id,
                     noise_augmentation_id,
                     noise_augmentation,
                     noise_offset) = _get_next_noise_offset(noise_file_id=noise_file_id,
                                                            noise_augmentation_id=noise_augmentation_id,
                                                            noise_offset=noise_offset,
                                                            target_length=target_length,
                                                            noise_files=noise_files,
                                                            noise_augmentations=noise_augmentations,
                                                            num_ir=num_ir)
                    used_noise_samples += target_length

                    used_noise_files.add(f'{noise_file_id}_{noise_augmentation_id}')

                    mixtures.append(Mixture(
                        targets=targets,
                        name=str(m_id),
                        noise=Noise(file_id=noise_file_id + 1,
                                    augmentation=noise_augmentation,
                                    offset=noise_offset),
                        samples=target_length,
                        snr=UniversalSNR(value=snr.value,
                                         is_random=snr.is_random),
                        spectral_mask_id=spectral_mask_id + 1,
                        spectral_mask_seed=randint(0, np.iinfo('i').max)))
                    m_id += 1

    return len(used_noise_files), used_noise_samples, mixtures


def _non_combinatorial_noise_mix(augmented_targets: AugmentedTargets,
                                 target_files: TargetFiles,
                                 target_augmentations: AugmentationRules,
                                 noise_files: NoiseFiles,
                                 noise_augmentations: AugmentationRules,
                                 spectral_masks: SpectralMasks,
                                 all_snrs: list[UniversalSNRGenerator],
                                 mixups: list[int],
                                 num_classes: int,
                                 truth_mutex: bool,
                                 feature_step_samples: int,
                                 num_ir: int) -> tuple[int, int, Mixtures]:
    """ Combine a target/augmentation with a single cut of a noise/augmentation non-exhaustively
    (each target/augmentation does not use each noise/augmentation). Cut has random start and loop back to
    beginning if end of noise/augmentation is reached.
    """
    from random import choice
    from random import randint

    import numpy as np

    from .datatypes import Mixture
    from .datatypes import Mixtures
    from .datatypes import Noise
    from .datatypes import UniversalSNR
    from .targets import get_augmented_target_ids_for_mixup

    mixtures: Mixtures = []
    m_id = 0
    used_noise_files = set()
    used_noise_samples = 0
    noise_file_id = None
    noise_augmentation_id = None

    augmented_target_indices_for_mixups = [get_augmented_target_ids_for_mixup(
        augmented_targets=augmented_targets,
        targets=target_files,
        target_augmentations=target_augmentations,
        mixup=mixup,
        num_classes=num_classes,
        truth_mutex=truth_mutex) for mixup in mixups]
    for mixup in augmented_target_indices_for_mixups:
        for augmented_target_indices in mixup:
            targets, target_length = _get_target_info(augmented_target_ids=augmented_target_indices,
                                                      augmented_targets=augmented_targets,
                                                      target_files=target_files,
                                                      target_augmentations=target_augmentations,
                                                      feature_step_samples=feature_step_samples,
                                                      num_ir=num_ir)

            for spectral_mask_id in range(len(spectral_masks)):
                for snr in all_snrs:
                    (noise_file_id,
                     noise_augmentation_id,
                     noise_augmentation,
                     noise_length) = _get_next_noise_indices(noise_file_id=noise_file_id,
                                                             noise_augmentation_id=noise_augmentation_id,
                                                             noise_files=noise_files,
                                                             noise_augmentations=noise_augmentations,
                                                             num_ir=num_ir)
                    used_noise_samples += target_length

                    used_noise_files.add(f'{noise_file_id}_{noise_augmentation_id}')

                    mixtures.append(Mixture(
                        targets=targets,
                        name=str(m_id),
                        noise=Noise(file_id=noise_file_id + 1,
                                    augmentation=noise_augmentation,
                                    offset=choice(range(noise_length))),
                        samples=target_length,
                        snr=UniversalSNR(value=snr.value,
                                         is_random=snr.is_random),
                        spectral_mask_id=spectral_mask_id + 1,
                        spectral_mask_seed=randint(0, np.iinfo('i').max)))
                    m_id += 1

    return len(used_noise_files), used_noise_samples, mixtures


def _get_next_noise_indices(noise_file_id: int,
                            noise_augmentation_id: int,
                            noise_files: NoiseFiles,
                            noise_augmentations: AugmentationRules,
                            num_ir: int) -> tuple[int, int, Augmentation, int]:
    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length

    if noise_file_id is None or noise_augmentation_id is None:
        noise_file_id = 0
        noise_augmentation_id = 0
    else:
        noise_augmentation_id += 1
        if noise_augmentation_id == len(noise_augmentations):
            noise_augmentation_id = 0
            noise_file_id += 1
            if noise_file_id == len(noise_files):
                noise_file_id = 0

    noise_augmentation = augmentation_from_rule(noise_augmentations[noise_augmentation_id], num_ir)
    noise_length = estimate_augmented_length_from_length(length=noise_files[noise_file_id].samples,
                                                         tempo=noise_augmentation.tempo)
    return noise_file_id, noise_augmentation_id, noise_augmentation, noise_length


def _get_next_noise_offset(noise_file_id: int | None,
                           noise_augmentation_id: int | None,
                           noise_offset: int | None,
                           target_length: int,
                           noise_files: NoiseFiles,
                           noise_augmentations: AugmentationRules,
                           num_ir: int) -> tuple[int, int, Augmentation, int]:
    from sonusai import SonusAIError
    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length

    if noise_file_id is None or noise_augmentation_id is None or noise_offset is None:
        noise_file_id = 0
        noise_augmentation_id = 0
        noise_offset = 0

    noise_augmentation = augmentation_from_rule(noise_augmentations[noise_file_id], num_ir)
    noise_length = estimate_augmented_length_from_length(length=noise_files[noise_file_id].samples,
                                                         tempo=noise_augmentation.tempo)
    if noise_offset + target_length >= noise_length:
        if noise_offset == 0:
            raise SonusAIError('Length of target audio exceeds length of noise audio')

        noise_offset = 0
        noise_augmentation_id += 1
        if noise_augmentation_id == len(noise_augmentations):
            noise_augmentation_id = 0
            noise_file_id += 1
            if noise_file_id == len(noise_files):
                noise_file_id = 0
        noise_augmentation = augmentation_from_rule(noise_augmentations[noise_augmentation_id], num_ir)

    return noise_file_id, noise_augmentation_id, noise_augmentation, noise_offset


def _get_target_info(augmented_target_ids: list[int],
                     augmented_targets: AugmentedTargets,
                     target_files: TargetFiles,
                     target_augmentations: AugmentationRules,
                     feature_step_samples: int,
                     num_ir: int) -> tuple[Targets, int]:
    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length
    from .datatypes import Target
    from .datatypes import Targets

    mixups: Targets = []
    target_length = 0
    for idx in augmented_target_ids:
        tfi = augmented_targets[idx].target_id
        target_augmentation_rule = target_augmentations[augmented_targets[idx].target_augmentation_id]
        target_augmentation = augmentation_from_rule(target_augmentation_rule, num_ir)

        mixups.append(Target(file_id=tfi + 1, augmentation=target_augmentation))

        target_length = max(estimate_augmented_length_from_length(length=target_files[tfi].samples,
                                                                  tempo=target_augmentation.tempo,
                                                                  frame_length=feature_step_samples),
                            target_length)
    return mixups, target_length


def get_all_snrs_from_config(config: dict) -> list[UniversalSNRGenerator]:
    from .datatypes import UniversalSNRGenerator

    return ([UniversalSNRGenerator(is_random=False, _raw_value=snr) for snr in config['snrs']] +
            [UniversalSNRGenerator(is_random=True, _raw_value=snr) for snr in config['random_snrs']])



import numpy as np


def calc_wsdr(hypothesis: np.ndarray,
              reference: np.ndarray,
              with_log: bool = False,
              with_negate: bool = False) -> tuple[float, np.ndarray, np.ndarray]:
    """Calculate weighted SDR (signal distortion ratio) using all source inputs of size [samples, nsrc].
       Uses true reference energy ratios to weight each cross-correlation coefficient cc = <y,yˆ>/∥y∥∥yˆ∥
       in a sum over all sources.

       range is -1 --> 1 as correlation/estimation improves or with_log -3db --> 70db (1e7 max)
       if with_negate, range is 1 --> -1 as correlation improves and with_log range 3db --> -70db (1e-7 min)

       Returns:  wsdr      scalar weighted signal-distortion ratio
                 ccoef     nsrc vector of cross correlation coefficients
                 cweights  nsrc vector of reference energy ratio weights

    Reference:
        WSDR: 2019-ICLR-dcunet-phase-aware-speech-enh

    :param hypothesis: [samples, nsrc]
    :param reference: [samples, nsrc]
    :param with_log: enable scaling (return 10*log10)
    :param with_negate: enable negation (for use as a loss function)
    :return: (wsdr, ccoef, cweights)
    """
    nsrc = reference.shape[-1]
    assert hypothesis.shape[-1] == nsrc

    # Calculate cc = <y,yˆ>/∥y∥∥yˆ∥ always in range -1 --> 1, size [1,nsrc]
    ref_e = np.sum(reference ** 2, axis=0, keepdims=True)  # [1,nsrc]
    hy_e = np.sum(hypothesis ** 2, axis=0, keepdims=True)
    allref_e = np.sum(ref_e)
    cc = np.zeros(nsrc)  # calc correlation coefficient
    cw = np.zeros(nsrc)  # cc weights (energy ratio)
    for i in range(nsrc):
        denom = np.sqrt(ref_e[0, i]) * np.sqrt(hy_e[0, i]) + 1e-7
        cc[i] = np.sum(reference[:, i] * hypothesis[:, i], axis=0, keepdims=True) / denom
        cw[i] = ref_e[0, i] / (allref_e + 1e-7)

    # Note: tests show cw sums to 1.0 (+/- 7 digits), so just use cw for weighted sum
    if with_negate:  # for use as a loss function
        wsdr = float(np.sum(cw * -cc))  # cc always in range 1 --> -1
        if with_log:
            wsdr = max(wsdr, -1.0)
            wsdr = 10 * np.log10(wsdr + 1 + 1e-7)  # range 3 --> -inf (or 1e-7 limit of -70db)
    else:
        wsdr = float(np.sum(cw * cc))  # cc always in range -1 --> 1
        if with_log:
            wsdr = min(wsdr, 1.0)  # (np.sum(cw * cc) needs sat ==1.0 for log)
            wsdr = 10 * np.log10(-1 / (wsdr - 1 - 1e-7))  # range -3 --> inf (or 1e-7 limit of 70db)

    return float(wsdr), cc, cw

    # From calc_sa_sdr:
    # These should include a noise to be a complete mixture estimate, i.e.,
    #     noise_est = sum-over-all-srcs(s_est(0:nsamples, :) - sum-over-non-noisesrc(s_est(0:nsamples, n))
    # should be one of the sources in reference (s_true) and hypothesis (s_est).
    #
    # Calculates -10*log10(sumn(||sn||^2) / sumn(||sn - shn||^2)
    # Note: for SA method, sums are done independently on ref and error before division, vs. SDR and SI-SDR
    # where sum over n is taken after divide (before log).  This is more stable in noise-only cases and also
    # when some sources are poorly estimated.
    # TBD: add soft-max option with eps and tau params
    #
    # if with_scale:
    #     # calc 1 x nsrc scaling factors
    #     ref_energy = np.sum(reference ** 2, axis=0, keepdims=True)
    #     # if ref_energy is zero, just set scaling to 1.0
    #     with np.errstate(divide='ignore', invalid='ignore'):
    #         opt_scale = np.sum(reference * hypothesis, axis=0, keepdims=True) / ref_energy
    #         opt_scale[opt_scale == np.inf] = 1.0
    #         opt_scale = np.nan_to_num(opt_scale, nan=1.0)
    #     scaled_ref = opt_scale * reference
    # else:
    #     scaled_ref = reference
    #     opt_scale = np.ones((1, reference.shape[1]), dtype=float)
    #
    # # Calculate Lsdr = −<y,yˆ>/∥y∥∥yˆ∥ always in range [1 --> -1], size [batch,]
    # t_tru_sq = torch.sum(torch.square(t_tru), -1)
    # t_denom = torch.sqrt(t_tru_sq) * torch.sqrt(torch.sum(torch.square(t_est), -1)) + 1e-7
    # t_wsdr = -torch.divide(torch.sum(torch.multiply(t_tru, t_est), -1), t_denom)
    # n_tru_sq = torch.sum(torch.square(n_tru), -1)
    # n_denom = torch.sqrt(torch.sum(torch.square(n_tru), -1)) \
    #           * torch.sqrt(torch.sum(torch.square(n_est), -1)) + 1e-7
    # n_wsdr = -torch.divide(torch.sum(torch.multiply(n_tru, n_est), -1), n_denom)
    # if self.cl_noise_wght > 0:
    #     wsdr = self.cl_target_wght * t_wsdr + self.cl_noise_wght * n_wsdr
    # else:  # adaptive per relative strength of target vs noise:  α = ||y||2/(||y||2 +||z||2)
    #     tweight = torch.divide(t_tru_sq, t_tru_sq + n_tru_sq + 1e-7)  # energy ratio target vs. noise
    #     wsdr = tweight * t_wsdr + (1 - tweight) * n_wsdr
    # wsdr = torch.mean(wsdr)  # reduction to scalar
    #
    # # multisrc sa-sdr, inputs must be [samples, nsrc]
    # err = scaled_ref - hypothesis
    #
    # # -10*log10(sumk(||sk||^2) / sumk(||sk - shk||^2)
    # # sum over samples and sources
    # num = np.sum(reference ** 2)
    # den = np.sum(err ** 2)
    # if num == 0 and den == 0:
    #     ratio = np.inf
    # else:
    #     ratio = num / (den + np.finfo(np.float32).eps)
    #
    # sa_sdr = 10 * np.log10(ratio)
    #
    # if with_negate:
    #     # for use as a loss function
    #     sa_sdr = -sa_sdr
    #
    # return sa_sdr, opt_scale



# SonusAI data generator classes
from .keras_from_mixdb import KerasFromH5
from .keras_from_mixdb import KerasFromMixtureDatabase
from .torch_from_mixdb import AawareDataLoader
from .torch_from_mixdb import TorchFromMixtureDatabase



import numpy as np
import pandas as pd

from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Predict
from sonusai.mixture import Truth


def confusion_matrix_summary(mixdb: MixtureDatabase,
                             mixids: GeneralizedIDs,
                             truth_f: Truth,
                             predict: Predict,
                             class_idx: int,
                             predict_thr: float | np.ndarray = 0,
                             truth_thr: float = 0.5,
                             timesteps: int = 0) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Calculate confusion matrix for specified class, using truth and prediction
       data [features, num_classes].

       predict_thr sets the decision threshold(s) applied to predict data, thus allowing
       predict to be continuous probabilities.

       Default predict_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
       if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
       the confusion matrix is calculated for all classes.

       Returns pandas dataframes of confusion matrix cmdf and normalized confusion matrix cmndf.
    """
    from sonusai.metrics import one_hot

    num_classes = truth_f.shape[1]
    # TODO: re-work for modern mixdb API
    ytrue, ypred = get_mixids_data(mixdb=mixdb, mixids=mixids, truth_f=truth_f, predict=predict)  # type: ignore

    # Check predict_thr array or scalar and return final scalar predict_thr value
    if not mixdb.truth_mutex and num_classes > 1:
        if not isinstance(predict_thr, np.ndarray):
            if predict_thr == 0:
                # multi-label predict_thr scalar 0 force to 0.5 default
                predict_thr = np.atleast_1d(0.5)
            else:
                predict_thr = np.atleast_1d(predict_thr)
        else:
            if predict_thr.ndim == 1:
                if predict_thr[0] == 0:
                    # multi-label predict_thr array scalar 0 force to 0.5 default
                    predict_thr = np.atleast_1d(0.5)
                else:
                    # multi-label predict_thr array set to scalar = array[0]
                    predict_thr = predict_thr[0]
            else:
                # multi-label predict_thr array scalar set = array[class_idx]
                predict_thr = predict_thr[class_idx]

    if len(mixdb.class_labels) == num_classes:
        class_names = mixdb.class_labels
    else:
        class_names = ([f'Class {i}' for i in range(1, num_classes + 1)])

    class_nums = ([f'{i}' for i in range(1, num_classes + 1)])

    if mixdb.truth_mutex:
        # single-label mode force to argmax mode
        predict_thr = np.array(0, dtype=np.float32)
        _, _, cm, cmn, _, _ = one_hot(ytrue, ypred, predict_thr, truth_thr, timesteps)
        row_n = class_names
        row_n[-1] = 'Other'
        # mux = pd.MultiIndex.from_product([['Single-label/mutex mode, truth thr = {}'.format(truth_thr)],
        #                                   class_nums])
        # mux = pd.MultiIndex.from_product([['truth thr = {}'.format(truth_thr)], class_nums])

        cmdf = pd.DataFrame(cm, index=row_n, columns=class_nums, dtype=np.int32)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=class_nums, dtype=np.float32)

    else:
        _, _, cm, cmn, _, _ = one_hot(ytrue[:, class_idx], ypred[:, class_idx], predict_thr, truth_thr, timesteps)
        cname = class_names[class_idx]
        row_n = ['TrueN', 'TrueP']
        col_n = ['N-' + cname, 'P-' + cname]
        cmdf = pd.DataFrame(cm, index=row_n, columns=col_n, dtype=np.int32)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=col_n, dtype=np.float32)
        # add thresholds in 3rd row
        pdnote = pd.DataFrame(np.atleast_2d([predict_thr, truth_thr]), index=['p/t thr:'], columns=col_n)
        cmdf = pd.concat([cmdf, pdnote])
        cmndf = pd.concat([cmndf, pdnote])

    return cmdf, cmndf



# SonusAI query utilities
from .queries import get_mixids_from_noise
from .queries import get_mixids_from_snr
from .queries import get_mixids_from_target
from .queries import get_mixids_from_truth_function
from .queries import get_mixids_from_truth_index



from sonusai.mixture.datatypes import Truth
from sonusai.mixture.truth_functions.data import Data


def _core(data: Data, mapped: bool, snr: bool) -> Truth:
    import numpy as np

    from sonusai import SonusAIError
    from sonusai.mixture import calculate_mapped_snr_f
    from sonusai.utils import compute_energy_f

    snr_db_mean = None
    snr_db_std = None
    if mapped:
        if data.config.config is None:
            raise SonusAIError('Truth function mapped SNR missing config')

        parameters = ['snr_db_mean', 'snr_db_std']
        for parameter in parameters:
            if parameter not in data.config.config:
                raise SonusAIError(f'Truth function mapped_snr_f config missing required parameter: {parameter}')

        snr_db_mean = data.config.config['snr_db_mean']
        if len(snr_db_mean) != data.target_fft.bins:
            raise SonusAIError(f'Truth function mapped_snr_f snr_db_mean does not have {data.target_fft.bins} elements')

        snr_db_std = data.config.config['snr_db_std']
        if len(snr_db_std) != data.target_fft.bins:
            raise SonusAIError(f'Truth function mapped_snr_f snr_db_std does not have {data.target_fft.bins} elements')

    for index in data.zero_based_indices:
        if index + data.target_fft.bins > data.config.num_classes:
            raise SonusAIError('Truth index exceeds the number of classes')

    target_energy = compute_energy_f(time_domain=data.target_audio, transform=data.target_fft)
    noise_energy = None
    if snr:
        noise_energy = compute_energy_f(time_domain=data.noise_audio, transform=data.noise_fft)

    if len(target_energy) != len(data.offsets):
        raise SonusAIError(f'Number of frames in target_energy, {len(target_energy)},'
                           f' is not number of frames in truth, {len(data.offsets)}')

    for idx, offset in enumerate(data.offsets):
        tmp = target_energy[idx]

        if snr:
            old_err = np.seterr(divide='ignore', invalid='ignore')
            tmp /= noise_energy[idx]
            np.seterr(**old_err)

        tmp = np.nan_to_num(tmp, nan=-np.inf, posinf=np.inf, neginf=-np.inf)

        if mapped:
            tmp = calculate_mapped_snr_f(tmp, snr_db_mean, snr_db_std)

        for index in data.zero_based_indices:
            data.truth[offset:offset + data.frame_size, index:index + data.target_fft.bins] = tmp

    return data.truth


def energy_f(data: Data) -> Truth:
    """Frequency domain energy truth generation function

Calculates the true energy per bin:

Ti^2 + Tr^2

where T is the target STFT bin values.

Output shape: [:, bins]
    """
    return _core(data=data, mapped=False, snr=False)


def snr_f(data: Data) -> Truth:
    """Frequency domain SNR truth function documentation

Calculates the true SNR per bin:

(Ti^2 + Tr^2) / (Ni^2 + Nr^2)

where T is the target and N is the noise STFT bin values.

Output shape: [:, bins]
    """
    return _core(data=data, mapped=False, snr=True)


def mapped_snr_f(data: Data) -> Truth:
    """Frequency domain mapped SNR truth function documentation

Output shape: [:, bins]
    """
    return _core(data=data, mapped=True, snr=True)


def energy_t(data: Data) -> Truth:
    """Time domain energy truth function documentation

Calculates the true time domain energy of each frame:

For OLS:
    sum(x[0:N-1]^2) / N

For OLA:
    sum(x[0:R-1]^2) / R

where x is the target time domain data,
N is the size of the transform, and
R is the number of new samples in the frame.

Output shape: [:, 1]

Note: feature transforms can be defined to use a subset of all bins,
i.e., subset of 0:128 for N=256 could be 0:127 or 1:128. energy_t
will reflect the total energy over all bins regardless of the feature
transform config.
    """
    import numpy as np

    from sonusai import SonusAIError

    _, target_energy = data.target_fft.execute_all(data.target_audio)
    if len(target_energy) != len(data.offsets):
        raise SonusAIError(f'Number of frames in target_energy, {len(target_energy)},'
                           f' is not number of frames in truth, {len(data.offsets)}')

    for offset in data.offsets:
        data.truth[offset:offset + data.frame_size, data.zero_based_indices] = np.float32(target_energy)

    return data.truth



"""sonusai doc

usage: doc [-h] [TOPIC]

options:
   -h, --help   Display this help.

Show SonusAI documentation.

"""


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    from sonusai import doc

    topic = args['TOPIC']

    print(f'SonusAI {sonusai.__version__} Documentation')
    print('')

    topics = sorted([item[4:] for item in dir(doc) if item.startswith('doc_')])

    if topic not in topics:
        if topic is not None:
            print(f'Unknown topic: {topic}')
            print('')

        print('Available topics:')
        for item in topics:
            print(f'  {item}')
        return

    text = getattr(doc, 'doc_' + topic)()
    print(text[1:])


if __name__ == '__main__':
    main()
